Package: LBBNN
Title: Latent Binary Bayesian Neural Networks Using 'torch'
Version: 0.1.6
Authors@R: c(person("Lars", "Skaaret-Lund", email = "lars.skaaret-lund@nmbu.no", role = c("aut", "cre")), person("Aliaksandr", "Hubin", email = "aliaksandr.hubin@nmbu.no", role = c("aut")), person("Eirik", "Høyheim", email = "eirik.hoyheim@ffi.no", role = "aut"))
Maintainer: Lars Skaaret-Lund <lars.skaaret-lund@nmbu.no>
Description: Latent binary Bayesian neural networks (LBBNNs) are implemented using 
    'torch', an R interface to the LibTorch backend. Supports mean-field variational 
    inference as well as flexible variational posteriors using normalizing flows. 
    The standard LBBNN implementation follows Hubin and Storvik (2024) <doi:10.3390/math12060788>, 
    using the local reparametrization trick as in Skaaret-Lund et al. (2024) 
    <https://openreview.net/pdf?id=d6kqUKzG3V>. Input-skip connections are also supported, 
    as described in Høyheim et al. (2025) <doi:10.48550/arXiv.2503.10496>.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.2
Language: en-US
Suggests: testthat (>= 3.0.0), knitr, rmarkdown, torchvision
Config/testthat/edition: 3
Depends: R (>= 3.5)
LazyData: true
VignetteBuilder: knitr
Imports: ggplot2, torch, igraph, coro, svglite
NeedsCompilation: no
Packaged: 2026-06-30 12:58:31 UTC; larsskaaret-lund
Author: Lars Skaaret-Lund [aut, cre],
  Aliaksandr Hubin [aut],
  Eirik Høyheim [aut]
Repository: CRAN
Date/Publication: 2026-06-30 14:00:02 UTC
Built: R 4.5.2; ; 2026-06-30 16:20:34 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(coef,lbbnn_net)
S3method(plot,lbbnn_net)
S3method(predict,lbbnn_net)
S3method(print,lbbnn_net)
S3method(residuals,lbbnn_net)
S3method(summary,lbbnn_net)
export(custom_activation)
export(get_dataloaders)
export(get_local_explanations_gradient)
export(lbbnn_conv2d)
export(lbbnn_linear)
export(lbbnn_net)
export(normalizing_flow)
export(quants)
export(resolve_device)
export(rnvp_layer)
export(torch_available)
export(train_lbbnn)
export(validate_lbbnn)
importFrom(torch,torch_empty)
importFrom(torch,torch_exp)
importFrom(torch,torch_long)
importFrom(torch,torch_rand)
importFrom(torch,torch_randn)
importFrom(torch,torch_zeros)
importFrom(torch,torch_zeros_like)
importFrom(torch,with_no_grad)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# LBBNN

<!-- badges: start -->

[![](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![](https://img.shields.io/github/last-commit/LarsELund/LBBNN.svg)](https://github.com/LarsELund/LBBNN/commits/main)
[![](https://img.shields.io/github/languages/code-size/LarsELund/LBBNN.svg)](https://github.com/LarsELund/LBBNN)
[![R-CMD-check](https://github.com/LarsELund/LBBNN/workflows/R-CMD-check/badge.svg)](https://github.com/LarsELund/LBBNN/actions)
[![License:
MIT](https://img.shields.io/badge/license-MIT-blue.svg)](LICENSE)

<!-- badges: end -->

The goal of LBBNN is to implement Latent Bayesian Binary Neural Networks
(<https://openreview.net/pdf?id=d6kqUKzG3V>) in R, using the torch for R
package. Currently, standard LBBNNs are implemented. In the future, we
will also implement LBBNNs with input-skip (see
<https://arxiv.org/abs/2503.10496>).

## Installation

You can install the development version of LBBNN from
[GitHub](https://github.com/LarsELund/LBBNN) with:

``` r
# install.packages("pak")
pak::pak("LarsELund/LBBNN")
```

## Example

This example demonstrates how to implement a simple feed forward LBBNN
on the raisin_dataset.

First, data must be processed and moved to torch dataloaders, using the
get_dataloaders() function:

``` r
library(LBBNN)
library(ggplot2)
library(torch)

#the get_dataloaders function takes a data.frame dataset as input, then splits the data
#in a training and validation set based on train_proportion, and returns torch dataloader 
#objects.
loaders <- get_dataloaders(raisin_dataset, train_proportion = 0.8, 
                           train_batch_size = 720, test_batch_size = 180)
train_loader <- loaders$train_loader
test_loader <- loaders$test_loader
```

To initialize the LBBNN, we first define some hyperparameters. First,
the user must define what type of problem they are facing. This could be
either binary classification (as in this case), multiclass
classification (more than two classes), or regression (continuous
output). Next, the user needs to define a size vector. The first element
is the number of variables in the dataset (7 in this case), the last
element is the number of output neurons (1 here), and the elements in
between represent the number of neurons in the hidden layer(s). Then,
the user must define the prior inclusion probability for each weight
matrix. This parameter is important as it reflects prior beliefs about
how dense the network should be. The user also needs to define the prior
standard deviation for the weight and bias parameters. Lastly, the user
must define the initialization of the inclusion parameters.

``` r
problem <- 'binary classification'
sizes <- c(7, 5, 5, 1) #7 input variables, two hidden layers of 5 neurons each, 1 output neuron.
inclusion_priors <-c(0.5, 0.5, 0.5) #one prior probability per weight matrix.
stds <- c(1, 1, 1) #prior standard deviation for each layer.
inclusion_inits <- matrix(rep(c(-10, 15), 3),nrow = 2, ncol = 3) #one low and high for each layer.
device <- 'cpu' #can also be mps or gpu.
```

We can now define the model. We use the mean-field posterior and
input-skip.

``` r
torch_manual_seed(0)
model_input_skip <- lbbnn_net(problem_type = problem,sizes = sizes,prior = inclusion_priors,
                      inclusion_inits = inclusion_inits,input_skip = TRUE,std = stds,
                   flow = FALSE,device = device)
```

To train the model, the function train_lbbnn is used.

``` r
results_input_skip <- suppressMessages(train_lbbnn(epochs = 800,LBBNN = model_input_skip, lr = 0.01,train_dl = train_loader,device = device))
#save the model 
#torch::torch_save(model_input_skip$state_dict(), 
#paste(getwd(),'/R/saved_models/README_input_skip_example_model.pth',sep = ''))
```

To evaluate performance on the validation data, the function
validate_lbbnn is used. It takes a model, number of samples for model
averaging, and the validation data as input.

``` r
validate_lbbnn(LBBNN = model_input_skip,num_samples = 100,test_dl = test_loader,device)
#> $accuracy_full_model
#> [1] 0.8666667
#> 
#> $accuracy_sparse
#> [1] 0.8555555
#> 
#> $density
#> [1] 0.1214953
#> 
#> $density_active_path
#> [1] 0.09345794
#validate_lbbnn(LBBNN = model_flows,num_samples = 1000,test_dl = test_loader,device)
```

Plot the global structure (explanation) for the given model:

``` r
plot(model_input_skip,type = 'global',vertex_size = 13,edge_width = 0.6,label_size = 0.6)
```

<img src="man/figures/README-unnamed-chunk-6-1.png" width="100%" />

Note that only 4 of the 7 input variables are used.

This can also be seen using the summary function:

``` r
summary(model_input_skip)
#> Summary of lbbnn_net object:
#> -----------------------------------
#> Shows the number of times each variable was included from each layer
#> -----------------------------------
#> Then the average inclusion probability for each input from each layer
#> -----------------------------------
#> The final column shows the average inclusion probability across all layers
#> -----------------------------------
#>    L0 L1 L2    a0    a1    a2 a_avg
#> x0  0  0  0 0.136 0.077 0.035 0.100
#> x1  0  0  0 0.325 0.064 0.149 0.190
#> x2  0  1  0 0.076 0.336 0.080 0.194
#> x3  1  0  1 0.425 0.267 0.999 0.406
#> x4  1  1  0 0.327 0.352 0.058 0.314
#> x5  0  1  0 0.140 0.280 0.230 0.212
#> x6  0  0  1 0.058 0.256 0.988 0.232
#> The model took 11.162 seconds to train, using cpu
```

Local explanations for a specific sample using plot():

``` r
x <- train_loader$dataset$tensors[[1]] #grab the dataset
ind <- 42
data <- x[ind,] #plot this specific data-point
plot(model_input_skip,type = 'local',data = data,num_samples = 100)
```

<img src="man/figures/README-unnamed-chunk-8-1.png" width="100%" />

Compute residuals: y_true - y_predicted

``` r
residuals(model_input_skip)[1:10]
#>  [1] -0.23637275  0.05227208 -0.04005997  0.13478029  0.20345378 -0.09961643
#>  [7]  0.05248320  0.14917880 -0.07054520  0.06071264
```

Get local explanations using coef():

``` r
coef(model_input_skip,dataset = train_loader,inds = c(2,3,4,5,6))
#>         lower       mean      upper
#> x0  0.0000000  0.0000000  0.0000000
#> x1  0.0000000  0.0000000  0.0000000
#> x2 -1.4081770 -1.0450520 -0.1182637
#> x3 -3.0491267 -1.5438678 -0.5097757
#> x4 -1.5836792 -0.5174002  0.8989937
#> x5  0.0000000  0.3023412  0.5179587
#> x6 -0.7162537 -0.6587479 -0.6052896
```

Get posterior predictions:

``` r
predictions <- predict(model_input_skip,mpm = TRUE,newdata = test_loader,draws = 100)
dim(predictions) #shape is (draws,samples,classes)
#> [1] 100 180   1
```

Print the model:

``` r
print(model_input_skip)
#> 
#> ========================================
#>           LBBNN Model Summary           
#> ========================================
#> 
#> Module Overview:
#>   - An `nn_module` containing 343 parameters.
#> 
#> ---------------- Submodules ----------------
#>   - layers               : nn_module_list  # 305 parameters
#>   - layers.0             : lbbnn_linear    # 115 parameters
#>   - layers.1             : lbbnn_linear    # 190 parameters
#>   - act                  : nn_leaky_relu   # 0 parameters
#>   - out_layer            : lbbnn_linear    # 38 parameters
#>   - out                  : nn_sigmoid      # 0 parameters
#>   - loss_fn              : nn_bce_loss     # 0 parameters
#> 
#> Model Configuration:
#>   - LBBNN with input-skip 
#>   - Optimized using variational inference without normalizing flows 
#> 
#> Priors:
#>   - Prior inclusion probabilities per layer:  0.5, 0.5, 0.5 
#>   - Prior std dev for weights per layer:     1, 1, 1 
#> 
#> =================================================================
```

params <-
list(eval = TRUE)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(LBBNN)
has_torch <- requireNamespace("torch", quietly = TRUE) &&
            torch::torch_is_installed()

## ----eval = has_torch---------------------------------------------------------
torch::torch_manual_seed(42)
loaders_gs <- get_dataloaders(gallstone_dataset, train_proportion = 0.70,
                              train_batch_size = 223, test_batch_size = 96,
                              standardize = TRUE, seed = 42)
train_loader_gs <- loaders_gs$train_loader
test_loader_gs <- loaders_gs$test_loader

## ----eval = has_torch---------------------------------------------------------
problem <- "binary classification"
sizes <- c(40, 16, 16, 16, 16, 1)
inclusion_priors <- c(0.5, 0.5, 0.5, 0.5, 0.5) 
stds <- c(1, 1, 1, 1, 1) 
inclusion_inits <- 'polarized_dense'
device <- "cpu"
model_gs <- lbbnn_net(problem_type = problem, sizes = sizes,
                     prior = inclusion_priors, flow = TRUE,
                     dims = c(10, 10, 10, 10), 
                     inclusion_inits = inclusion_inits,
                     input_skip = TRUE, std = stds, device = device)
#print(model_gs)

## ----eval = FALSE-------------------------------------------------------------
# train_lbbnn(epochs = 1000, LBBNN = model_gs,
#             lr = 0.005, train_dl = train_loader_gs, device = device,
#             verbose = FALSE)
# 
# validate_lbbnn(LBBNN = model_gs, num_samples = 10, test_dl = test_loader_gs,
#               device = device)
# 

## ----eval = FALSE-------------------------------------------------------------
# torch::torch_manual_seed(42)
# model_2 <- lbbnn_net(problem_type = problem, sizes = sizes,
#                      prior = inclusion_priors, flow = TRUE,
#                      dims = c(10, 10, 10, 10),
#                      inclusion_inits = inclusion_inits,
#                      input_skip = TRUE, std = stds, device = device)
# 
# train_lbbnn(epochs = 1000, LBBNN = model_2,
#             lr = 0.005, train_dl = train_loader_gs, device = device,
#             verbose = FALSE, min_density = 0.1)
# 
# validate_lbbnn(LBBNN = model_2, num_samples = 10, test_dl = test_loader_gs,
#               device = device)

params <-
list(eval = TRUE)

## ----include=FALSE------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(LBBNN)
has_torch <- requireNamespace("torch", quietly = TRUE) &&
            torch::torch_is_installed()

## ----eval = has_torch---------------------------------------------------------
torch::torch_manual_seed(42)
loaders <- get_dataloaders(raisin_dataset, train_proportion = 0.8,
                           train_batch_size = 720, test_batch_size = 180)
train_loader <- loaders$train_loader
test_loader  <- loaders$test_loader

## ----eval = has_torch---------------------------------------------------------
problem <- "binary classification"
sizes <- c(7, 5, 5, 1)
inclusion_priors <- c(0.5, 0.5, 0.5)
stds <- c(1, 1, 1)
inclusion_inits <- 'balanced'
device <- "cpu"
model <- lbbnn_net(problem_type = problem, sizes = sizes,
                   prior = inclusion_priors,
                   inclusion_inits = inclusion_inits,
                   input_skip = TRUE, std = stds,
                   flow = FALSE, device = device)

## ----eval = has_torch---------------------------------------------------------
train_lbbnn(epochs = 10, LBBNN = model,
            lr = 0.05, train_dl = train_loader,
            device = device, verbose = FALSE)

## ----eval = has_torch---------------------------------------------------------
validate_lbbnn(LBBNN = model, num_samples = 2,
               test_dl = test_loader, device = device)

## ----fig.width=6, fig.height=6, eval = has_torch------------------------------
plot(model, type = 'global', vertex_size = 10, edge_width = 0.6, label_size = 0.6)

## ----fig.width=6, fig.height=6, eval = has_torch------------------------------
x_data <- train_loader$dataset$tensors[[1]] 
data <- x_data[42, ]
plot(model, type = "local", data = data,num_samples = 10)

## ----eval = has_torch---------------------------------------------------------
print(coef(model, data,num_samples = 10))

params <-
list(eval = TRUE)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(LBBNN)
has_torch <- requireNamespace("torch", quietly = TRUE) &&
            torch::torch_is_installed()

## ----eval = has_torch---------------------------------------------------------
i <- 1000
j <- 15
set.seed(42)
torch::torch_manual_seed(42)
X <- matrix(rnorm(i * j, mean = 0, sd = 1), ncol = j)
y_base <- c()
y_base <-  0.6 * X[, 1] - 0.4 * X[, 2] + 0.5 * X[, 3] + rnorm(n = i, sd = 0.1)
sim_data <- as.data.frame(X)
sim_data <- cbind(sim_data, y_base)
loaders <- get_dataloaders(sim_data, train_proportion = 0.9,
                           train_batch_size = 450, test_batch_size = 100,
                           standardize = FALSE)
train_loader <- loaders$train_loader
test_loader  <- loaders$test_loader

## ----eval = has_torch---------------------------------------------------------
problem <- "regression"
sizes <- c(j, 5, 5, 1) 
incl_priors <- c(0.5, 0.5, 0.5) 
stds <- c(1, 1, 1)
incl_inits <- 'polarized'
device <- "cpu" 
model_linear <- lbbnn_net(problem_type = problem, sizes = sizes,
                              prior = incl_priors, inclusion_inits = incl_inits,
                              std = stds, input_skip = TRUE, flow = FALSE,
                              num_transforms = 2, dims = c(10, 10, 10),
                              raw_output = FALSE, custom_act = NULL,
                              link = NULL, nll = NULL,
                              bias_inclusion_prob = FALSE, device = device)

## ----eval = has_torch---------------------------------------------------------
train_lbbnn(epochs = 50, LBBNN = model_linear,
            lr = 0.1, train_dl = train_loader, device = device, verbose = FALSE)
validate_lbbnn(LBBNN = model_linear, num_samples = 2, test_dl = test_loader,
              device = device)

## ----eval = has_torch---------------------------------------------------------
coef(model_linear, dataset = train_loader, inds = NULL,
     output_neuron = 1, num_data = 5, num_samples = 10)

## ----fig.width=6, fig.height=6, eval = has_torch------------------------------
plot(model_linear, type = "global", vertex_size = 7,
     edge_width = 0.4, label_size = 0.4)

#' @title Generate a custom activation function.
#' @description The first 3 entries are customized
#' in order to see if we can learn that structure. The rest will be relu as usual 
#' @return Returns a `\code{torch::nn_module} that can be used in an \code{LBBNN_Net}  
#' @export
custom_activation <- nn_module(
  "Custom_activation",
  initialize = function() {
    self$act <- torch::nn_leaky_relu(0.0)
    
  },
  forward = function(x) {
    shapes <- dim(x)
    x1 <- x[,1]
    x2 <- x[,2]
    x3 <- x[,3]
    x4 <- x[,4:shapes[2]]
    x1 <- torch::torch_exp(x1)$unsqueeze(2)
    x2 <- torch::torch_log(torch::torch_abs(x2))$unsqueeze(2)
    x3 <- torch::torch_abs(x3)^(1/3)
    x4 <- self$act(x4)
    
    x <- torch::torch_cat(c(x1,x2,x3,x4),dim = 2)
    return(x)
  }
)

#' @title Multi layer-perceptron
#' @description Generate a multi-layer perceptron, used in RNVP transforms.
#' @keywords internal
#' @return Returns a \code{torch::nn_module} representing the mlp.
#' The module has the following method:
#' \describe{
#'   \item{\code{forward(x)}}{
#'     Applies each linear layer in \code{hidden_sizes} followed by a LeakyReLU
#'     activation (except after
#'     the final layer). Returns a \code{torch::torch_tensor}
#'     whose last dimension equals the
#'     last element of \code{hidden_sizes}.
#'   }
#' }
mlp <- torch::nn_module(
  "mlp",
  initialize = function(hidden_sizes, device = "cpu") {
    self$lay <- torch::nn_module_list()

    for (i in 1:(length(hidden_sizes) - 1)) {
      self$lay$append(torch::nn_linear(hidden_sizes[i], hidden_sizes[i + 1]))
      if (i < length(hidden_sizes) - 1) {
        self$lay$append(torch::nn_leaky_relu()) #relu after each layer except the last
      }
    }
  },
  forward = function(x) {
    for (l in self$lay$children) {
      x <- l(x)
    }
    return(x)
  }
)

#' @title Single RNVP transform layer.
#' @param hidden_sizes A vector of integers.
#' The first is the dimensionality of the vector,
#' to be transformed by RNVP. The subsequent are hidden dimensions in the mlp.
#' @param device The device to be used. Default is CPU.
#' @description Affine half flow aka Real Non-Volume Preserving
#' (x = z * exp(s) + t), where a randomly selected half z1 of the dimensions
#' in z are transformed as an Affine function of the other half z2,
#' i.e. scaled by s(z2) and shifted by t(z2).
#' From "Density estimation using Real NVP", Dinh et al. (May 2016)
#' https://arxiv.org/abs/1605.08803
#' This implementation uses the numerically stable updates introduced by IAF:
#' https://arxiv.org/abs/1606.04934
#' @return
#' A \code{torch::nn_module} object representing a single RNVP layer.
#' The module has the following methods:
#' \describe{
#'   \item{\code{forward(z)}}{Applies the RNVP transformation.
#'   Returns a \code{torch::torch_tensor} with the
#'   same shape as z.}
#'   \item{\code{log_det()}}{A scalar \code{torch::torch_tensor}
#'   giving the log-determinant of the Jacobian of the transformation.}
#' }
#' @examples
#' \donttest{
#' if (torch_available()) {
#' z <- torch::torch_rand(200)
#' layer <- rnvp_layer(c(200,50,100))
#' out <- layer(z)
#' print(dim(out))
#' print(layer$log_det())}
#' }
#' @export
rnvp_layer <- torch::nn_module(
  "rnvp_layer",

  initialize = function(hidden_sizes, device = "cpu") {
    self$net <- mlp(hidden_sizes,device)
    self$t <- torch::nn_linear(hidden_sizes[length(hidden_sizes)],
                               hidden_sizes[1])
    self$s <- torch::nn_linear(hidden_sizes[length(hidden_sizes)],
                               hidden_sizes[1])
  },
  forward = function(z) {
    self$m <- torch::torch_bernoulli(0.5 * torch::torch_ones_like(z))
    z1 <- (1 - self$m) * z
    z2 <- self$m * z
    out <- self$net(z2)
    shift <- self$t(out)
    scale <- self$s(out)
    self$gate <- torch::torch_sigmoid(scale)
    x <- z1 * (self$gate + (1 - self$gate) * shift) + z2
    return(x)
  },
  log_det = function() {
    return(torch::torch_sum((1 - self$m) * torch::torch_log(self$gate + 1e-10)))
  }
)

#' @title Feed-forward Latent Binary Bayesian Neural Network (LBBNN)
#' @description
#' Each layer is defined by \code{lbbnn_linear}.
#' For example, \code{sizes = c(20, 200, 200, 5)} generates a network with:
#' \itemize{
#'   \item 20 input features,
#'   \item two hidden layers of 200 neurons each,
#'   \item an output layer with 5 neurons.
#'}
#' @param problem_type character, one of:
#' \code{'binary classification'}, \code{'multiclass classification'},
#' \code{'regression'}, or \code{'custom'}.
#' @param sizes Integer vector specifying the layer sizes of the network.
#' The first element is the input size,
#' the last is the output size,
#' and the intermediate integers represent hidden layers.
#' @param prior numeric vector of prior inclusion probabilities for
#' each weight matrix. length must be \code{length(sizes) - 1}.
#' @param std numeric vector of prior standard deviation for each weight matrix.
#' length must be \code{length(sizes) - 1}.
#' @param inclusion_inits numeric matrix of shape (2, number of weight matrices)
#' specifying the lower and upper bounds for initializations
#' of the inclusion parameters.
#' @param input_skip logical, whether to include input_skip.
#' @param flow logical, whether to use normalizing flows.
#' @param num_transforms integer, how many transformations to use in the flow.
#' @param dims numeric vector, hidden dimension for the neural network
#' in the RNVP transform.
#' @param device the device to be trained on. Can be 'cpu', 'gpu' or 'mps'.
#' Default is cpu.
#' @param raw_output logical, whether the network skips the last sigmoid/softmax
#'  layer to compute local explanations.
#' @param custom_act Allows the user to submit their own customized
#' activation function.
#' @param link User can define their own link function (not implemented).
#' @param nll User can define their own likelihood function (not implemented).
#' @param bias_inclusion_prob logical, determines whether the bias should
#' be as associated with inclusion probabilities.
#' @examples
#' \donttest{
#' if (torch_available()) {
#' layers <- c(10,2,5)
#' alpha <- c(0.3,0.9)
#' stds <- c(1.0,1.0)
#' inclusion_inits <- matrix(rep(c(-10,10),2),nrow = 2,ncol = 2)
#' prob <- 'multiclass classification'
#' net <- lbbnn_net(problem_type = prob, sizes = layers, prior = alpha,
#' std = stds, inclusion_inits = inclusion_inits,input_skip = FALSE,
#' flow = FALSE,device = 'cpu')
#' x <- torch::torch_rand(20,10,requires_grad = FALSE)
#' output <- net(x)
#' net$kl_div()$item()
#' net$density()}
#' }
#' @return A \code{torch::nn_module} object representing the LBBNN.
#'   It includes the following methods:
#'   \itemize{
#'     \item \code{forward(x, MPM = FALSE)}: Performs a forward pass
#'     through the whole network.
#'     \item \code{kl_div()}: Returns the KL divergence of the network.
#'     \item \code{density()}: Returns the density of the whole network,
#'     i.e. the proportion of weights
#'     with inclusion probabilities greater than 0.5.
#'     \item \code{compute_paths()}: Computes active paths through the network
#'     without input-skip.
#'     \item \code{compute_paths_input_skip()}: Computes active paths with
#'     input-skip enabled.
#'     \item \code{density_active_path()}: Returns network density after
#'     removing inactive paths.
#'   }
#'
#' @export
lbbnn_net <- torch::nn_module(
  "lbbnn_net",

  initialize = function(problem_type, sizes, prior, std, inclusion_inits,
                        input_skip = FALSE, flow = FALSE, num_transforms = 2,
                        dims = c(200, 200), device = "cpu", raw_output = FALSE,
                        custom_act = NULL, link = NULL, nll = NULL,
                        bias_inclusion_prob = FALSE) {
    self$device <- device
    self$layers <- torch::nn_module_list()
    self$problem_type <- problem_type
    self$input_skip <- input_skip
    self$flow <- flow
    self$bias_inclusion_prob <- bias_inclusion_prob
    self$num_transforms <- num_transforms
    self$dims <- dims
    self$sizes <- sizes
    self$prior_inclusion <- prior
    self$prior_std <- std
    self$elapsed_time <- 0 #to check how much time the model takes to train
    self$raw_output <- raw_output # TRUE when we want local explanations
    self$act <- torch::nn_leaky_relu(0.00)
    self$computed_paths <- FALSE
    if (! is.null(custom_act)) {
      self$act <- custom_act
      }
    if (length(prior) != length(sizes) - 1)
    (stop("Must have one prior inclusion probability per weight matrix"))

     #define the layers of the network
     #for the first layer input-skip and LBBNNs are the same,
     #but for all subsequent layers, input-skip will have shape (dim + p,dim)
     #where p is the number of variables. (i.e. the first element of sizes)
     for (i in 1:(length(sizes) - 2)) {
      if (i == 1 || input_skip == FALSE) {
        in_shape <- sizes[i]
        }
      else (in_shape <- sizes[i] + sizes[1])
      self$layers$append(lbbnn_linear(
        in_shape,
        sizes[i + 1],
        prior_inclusion = prior[i],
        standard_prior = std[i],
        density_init = inclusion_inits[, i],
        flow = self$flow,
        num_transforms = self$num_transforms,
        hidden_dims = self$dims,
        device = self$device,
        bias_inclusion_prob = self$bias_inclusion_prob))
    }
    if (input_skip) {
      out_size <- sizes[length(sizes) - 1] + sizes[1]
    } else {
      out_size <- sizes[length(sizes) - 1]
      }
    self$out_layer <- (lbbnn_linear(
      out_size,
      sizes[length(sizes)],
      prior_inclusion = prior[length(prior)],
      standard_prior = std[length(std)],
      density_init = inclusion_inits[, ncol(inclusion_inits)],
      flow = self$flow,
      num_transforms = self$num_transforms,
      hidden_dims = self$dims,
      device = self$device,
      bias_inclusion_prob = self$bias_inclusion_prob))

    if (problem_type == "binary classification") {
      self$out <- torch::nn_sigmoid()
      self$loss_fn <- torch::nn_bce_loss(reduction = "sum")
    }
    else if (problem_type == "multiclass classification" | problem_type == "MNIST") {
      self$out <- torch::nn_log_softmax(dim = 2)
      self$loss_fn <- torch::nn_nll_loss(reduction = "sum")
    }
    else if (problem_type == "regression") {
      self$out <- torch::nn_identity()
      self$loss_fn <- torch::nn_mse_loss(reduction = "sum")
    }else if (problem_type == "custom") {
      if (length(link) == 0 | length(nll) == 0)
        stop("Under custom problem, link function and the negative log likelihood
             must be provided as torch functions")
      self$out <- link()
      self$loss_fn <- nll(reduction = "sum")
    }
    else(stop("the type of problem must either be: \'binary classification\',
              \'multiclass classification\', \'regression\' or \'custom\'"))
  },
  forward = function(x, MPM = FALSE) {
    if (self$problem_type == "MNIST")(x <- x$view(c(-1, 28 * 28)))
    #regular LBBNN
    if (!self$input_skip) {
      x <- x$view(c(-1, self$sizes[1]))
      for (l in self$layers$children){
       x <- self$act(l(x, MPM)) #iterate over hidden layers
      }
      x <- self$out(self$out_layer(x, MPM))
    }
    #with input-skip
    else {x_input <- x$view(c(-1, self$sizes[1]))
    x <- self$act(self$layers$children$`0`(x_input, MPM)) #first layer
    j <- 1
    for (l in self$layers$children) {
      if (j > 1) {#skip the first layer when iterating.
        x <- self$act(l(torch::torch_cat(c(x, x_input), dim = 2), MPM))
      }
      j <- j + 1
    }

    #if we only want the raw output, skip sigmoid/softmax
    if (self$raw_output) {
      x <- self$out_layer(torch::torch_cat(c(x, x_input), dim = 2), MPM)
    }
    else (x <- self$out(self$out_layer(torch::torch_cat(c(x, x_input), dim = 2), MPM)))

    }

    return(x)
    },
  kl_div = function() {
    kl <- 0
    for (l in self$layers$children)(kl <- kl + l$kl_div())
    kl <- kl + self$out_layer$kl_div()
    return(kl)
  },
  compute_paths = function() {

    if (self$input_skip == TRUE) {
      stop("self$input_skip must be FALSE to use this function")
    }
    self$computed_paths <- TRUE
    #sending a random input through the network of alpha matrices (0 and 1)
    #and then backpropagating to find active paths
    a <- rnorm(n = self$layers$children$`0`$alpha$shape[2])
    x0 <- torch::torch_tensor(a, dtype = torch::torch_float32(), 
                              device = self$device)
    alpha_mats <- list() #initialize empty list to append network alphas

    for (l in self$layers$children) {
      lamd <- l$lambda_l$detach()
      alpha <- (torch::torch_sigmoid(lamd) > 0.5) * 1
      alpha$requires_grad <- TRUE
      alpha_mats <- append(alpha_mats, alpha)
      x0 <- torch::torch_matmul(x0, torch::torch_t(alpha))
    }
    #output layer
    lamd_out <- self$out_layer$lambda_l$detach()
    alpha_out <- (torch::torch_sigmoid(lamd_out) > 0.5) * 1
    alpha_out <- alpha_out$detach()
    alpha_out$requires_grad <- TRUE
    alpha_mats <- append(alpha_mats, alpha_out)
    x0 <- torch::torch_matmul(x0, torch::torch_t(alpha_out))
    L <- torch::torch_sum(x0) #summing in case more than 1 output. This is
    #equivalent to backpropagate for each output node.
    L$backward() #compute derivatives to get active paths
                  #any alpha preceding an alpha with value 0 will also become
                  #zero when gradients are passed backwards, and thus we will
                  #be left with the active paths.
    i <- 1
    alpha_mats_out <- list()
    for (j in self$layers$children) {
      alp <- alpha_mats[[i]] * alpha_mats[[i]]$grad
      alp[alp != 0] <- 1
      alpha_mats_out <- append(alpha_mats_out, alp$detach())
      j$alpha_active_path <- alp$detach()
      i <- i + 1

    }
    alp_out <- alpha_mats[[length(alpha_mats)]] * alpha_mats[[length(alpha_mats)]]$grad #last alpha
    alp_out[alp_out != 0] <- 1
    alpha_mats_out <- append(alpha_mats_out, alp_out$detach())
    self$out_layer$alpha_active_path <- alp_out$detach()
    return(alpha_mats_out)
  },
  compute_paths_input_skip = function() {
    if (self$input_skip == FALSE) {
      stop("self$input_skip must be TRUE to use this funciton")
    }
    self$computed_paths <- TRUE
    #sending a random input through the network of alpha matrices (0 and 1)
    #and then backpropagating to find active paths
    a <- rnorm(n = self$layers$children$`0`$alpha$shape[2])
    x0 <- torch::torch_tensor(a, dtype = torch::torch_float32(),
                              device = self$device)$unsqueeze(dim = 1)
    alpha_mats <- list() #initialize empty list to append network alphas
    lamd_input <- self$layers$children$`0`$lambda_l
    alpha_input <- (torch::torch_sigmoid(lamd_input) > 0.5) * 1
    alpha_input$requires_grad <- TRUE
    alpha_mats <- append(alpha_mats, alpha_input)
    x <- torch::torch_matmul(x0, torch::torch_t(alpha_input))
    j <- 1
    for (l in self$layers$children) {
      if (j > 1) {#skip the first layer when iterating.
        x <- (torch::torch_cat(c(x, x0), dim = 2))
        lamd <- l$lambda_l$detach()
        alpha <- (torch::torch_sigmoid(lamd) > 0.5) * 1
        alpha$requires_grad <- TRUE
        alpha_mats <- append(alpha_mats, alpha)
        x <- torch::torch_matmul(x, torch::torch_t(alpha))
      }
      j <- j + 1
    }
    #output layer
    lamd_out <- self$out_layer$lambda_l$detach()
    alpha_out <- (torch::torch_sigmoid(lamd_out) > 0.5) * 1
    alpha_out <- alpha_out$detach()
    alpha_out$requires_grad <- TRUE
    alpha_mats <- append(alpha_mats, alpha_out)
    x_out <- (torch::torch_cat(c(x, x0), dim = 2))
    x_out <- torch::torch_matmul(x_out, torch::torch_t(alpha_out))
    L <- torch::torch_sum(x_out) #summing in case more than 1 output. This is
    #equivalent to backpropagate for each output node.
    L$backward() #compute derivatives to get active paths
    #any alpha preceding an alpha with value 0 will also become
    #zero when gradients are passed backwards, and thus we will
    #be left with the active paths.
    i <- 1
    alpha_mats_out <- list()
    for (j in self$layers$children) {
      alp <- alpha_mats[[i]] * alpha_mats[[i]]$grad
      alp[alp != 0] <- 1
      alpha_mats_out <- append(alpha_mats_out, alp$detach())
      j$alpha_active_path <- alp$detach()
      i <- i + 1
    }
    alp_out <- alpha_mats[[length(alpha_mats)]] * alpha_mats[[length(alpha_mats)]]$grad #last alpha
    alp_out[alp_out != 0] <- 1
    alpha_mats_out <- append(alpha_mats_out, alp_out$detach())
    self$out_layer$alpha_active_path <- alp_out$detach()
    return(alpha_mats_out)
  },

  density = function() { #the standard density, without active paths
    alphas <- NULL
    for (l in self$layers$children) {
      alphas <- c(alphas, as.numeric(l$alpha$clone()$detach()))
      }
    alphas <- c(alphas, as.numeric(self$out_layer$alpha$clone()$detach()))
    return(mean(alphas > 0.5))


  },
  density_active_path = function() {#density when removing connections not within active paths
    num_incl <- 0
    tot <- 0
    paths <- c() #intialize empty vector to compute path depths (only for input-skip)
    p <- self$sizes[1]
    for (l in self$layers$children) {
      tot <- tot +  l$alpha_active_path$numel() #total number of alphas
      num_incl <- num_incl + torch::torch_sum(l$alpha_active_path) #number of 1s
      if (self$input_skip) {
        alp <- l$alpha_active_path
        count <- (alp[, (dim(alp)[2] - p + 1):dim(alp)[2]] != 0)$flatten()$sum()$item() #count number non-zero elements in the p last entries of alpha matrix. this gives active paths
        paths <- append(paths, count)
      }
    }
    num_incl <- num_incl + torch::torch_sum(self$out_layer$alpha_active_path) #output layer
    tot <- tot +  self$out_layer$alpha_active_path$numel()
  
    #if no input skip then avg and max depth is always equal to the number of layers
    avg_path_length <- length(self$layers) + 1
    max_path_length <- length(self$layers) + 1

    if (self$input_skip) {
      alp_out <- self$out_layer$alpha_active_path
      out_count <- (alp_out[, (dim(alp_out)[2] - p + 1):dim(alp_out)[2]] != 0)$flatten()$sum()$item()
      paths <- append(paths, out_count)
      count_vector <- c((length(self$layers) + 1):1) #from number of layers -> 1. number of layers is the maximum possible path length.
      avg_path_length <- (torch::torch_dot(paths, count_vector) / sum(paths))$item()
      max_path_length <- count_vector[torch::torch_argmax((paths != 0) * 1)$item()] #get the index of of the first non-zero item in count vector, corresponding to the maximum path
    }

    return(num_incl$item() / tot)

  }

)

#' @title Train an instance of \code{lbbnn_net}.
#' @description Function that for each epoch iterates through each mini-batch,
#' computing the loss and using back-propagation to update network parameters.
#' @param epochs integer, total number of epochs to train for, where one epoch
#' is a pass through the entire training dataset (all mini batches).
#' @param LBBNN An instance of  \code{lbbnn_net}, to be trained.
#' @param lr numeric, the learning rate to be used in the Adam optimizer.
#' @param train_dl An instance of \code{torch::dataloader}
#' consisting of a tensor dataset with features and targets.
#' @param device the device to be trained on. Default is 'cpu',
#' also accepts 'gpu' or 'mps'.
#' @param scheduler A torch learning rate scheduler object.
#' Can be used to decay learning rate for better convergence,
#' currently only supports 'step'.
#' @param sch_step_size Where to decay if using \code{torch::lr_step}.
#'  E.g. 1000 means learning rate is decayed every 1000 epochs.
#' @return a list containing the losses and accuracy (if classification)
#' and density for each epoch during training.
#' For comparisons sake we show the density with and without active paths.
#' @examples
#' \donttest{
#' if (torch_available()) {
#'x<-torch::torch_randn(3,2)
#'b <- torch::torch_rand(2)
#'y <- torch::torch_matmul(x,b)
#'train_data <- torch::tensor_dataset(x,y)
#'train_loader <- torch::dataloader(train_data,batch_size = 3,shuffle=FALSE)
#'problem<-'regression'
#'sizes <- c(2,1,1)
#'inclusion_priors <-c(0.9,0.2)
#'inclusion_inits <- matrix(rep(c(-10,10),2),nrow = 2,ncol = 2)
#'stds <- c(1.0,1.0)
#'model <- lbbnn_net(problem,sizes,inclusion_priors,stds,inclusion_inits,
#'flow = FALSE)
#'output <- train_lbbnn(epochs = 1,LBBNN = model, lr = 0.01,
#'train_dl = train_loader)}
#'}
#' @return A list with elements (returned invisibly):
#'   \describe{
#'     \item{accs}{Vector of accuracy per epoch (classification only).}
#'     \item{loss}{Vector of average loss per epoch.}
#'     \item{density}{Vector of network densities per epoch.}
#'   }
#'@export
train_lbbnn <- function(epochs, LBBNN, lr, train_dl, device = "cpu",
                        scheduler = NULL, sch_step_size = NULL) {
  opt <- torch::optim_adam(LBBNN$parameters, lr = lr)
  accs <- c()
  losses <- c()
  density <- c()
  out_layer_density <- c()
  active_path_dens <- c()
  if (!is.null(scheduler)) {
    if (scheduler == "step") {
      sl <- torch::lr_step(opt, step_size = sch_step_size, gamma = 0.1)
    }
  }
  LBBNN$elapsed_time <- 0
  start <- base::proc.time()
  for (epoch in 1:epochs) {
  #  if(LBBNN$input_skip){LBBNN$compute_paths_input_skip()}
  #  else(LBBNN$compute_paths)
    if (epoch == epochs) { #only need these at the last epoch for residuals
     LBBNN$y <- c()
     LBBNN$r <- c()
    }
    LBBNN$train()
    corrects <- 0
    totals <- 0
    train_loss <- c()
    # use coro::loop() for stability and performance
    coro::loop(for (b in train_dl) {

      opt$zero_grad()
      data <- b[[1]]$to(device = device)
      output <- LBBNN(data, MPM = FALSE)
      target <- b[[2]]$to(device = device)
      if (epoch == epochs) { #add the targets and outputs to y and r
        LBBNN$y <- c(LBBNN$y, as.numeric(target$clone()$detach()$cpu()))
        LBBNN$r <- c(LBBNN$r, as.numeric(output$clone()$detach()$squeeze()$cpu()))
        }
      if (LBBNN$problem_type == "multiclass classification" | LBBNN$problem_type == "MNIST") { #nll loss needs float tensors but bce loss needs long tensors
        target <- torch::torch_tensor(target, dtype = torch::torch_long())
      }
      else (output <- output$squeeze()) #remove last dimension from binary classifiction or regression
      loss <- LBBNN$loss_fn(output, target) + LBBNN$kl_div() / length(train_dl)

      if (LBBNN$problem_type == "multiclass classification" | LBBNN$problem_type == "MNIST") {
        prediction <- max.col(output)
        corrects <- corrects + sum(prediction == target)
        totals <- totals + length(target)
        train_loss <- c(train_loss, loss$item())
      }

      else if (LBBNN$problem_type == "binary classification") {
        corrects <- corrects + sum((output > 0.5) == target)
        totals <- totals + length(target)
        train_loss <- c(train_loss, loss$item())

      }
      else if (LBBNN$problem_type == "custom") {
        train_loss <- c(train_loss, loss$item())
      }
      else {#for regression
        train_loss <- c(train_loss, loss$item())

      }
      loss$backward()
      opt$step()

    })
    if (!is.null(scheduler)) {
      sl$step()
      }

    train_acc <- corrects / totals
    if (LBBNN$problem_type != "regression") {
      message(sprintf(
        "\nEpoch %d, training: loss = %3.5f, acc = %3.5f, density = %3.5f",
        epoch, mean(train_loss), train_acc, LBBNN$density()
      ))
      accs <- c(accs, train_acc$item())
      losses <- c(losses, mean(train_loss))
    }
    if (LBBNN$problem_type == "regression") {
      message(sprintf(
        "\nEpoch %d, training: loss = %3.5f, density = %3.5f \n",
        epoch, mean(train_loss), LBBNN$density()
      ))
      losses <- c(losses, mean(train_loss))
    }
    density <- c(density, LBBNN$density())
  }
  l <- list("accs" = accs, "loss" = losses, "density" = density)
  time <- base::proc.time() - start
  LBBNN$elapsed_time <- time[[3]]
  invisible(l)
}


#' @title Validate a trained LBBNN model.
#' @description Computes metrics on a validation dataset without computing gradients.
#' Supports model averaging (recommended) by
#' sampling from the variational posterior (\code{num_samples} > 1)
#' to improve predictions. Returns metrics for both the full model
#' and the sparse model.
#' @param LBBNN An instance of a trained \code{lbbnn_net} to be validated.
#' @param num_samples integer, the number of samples from the variational
#' posterior to be used for model averaging.
#' @param test_dl An instance of \code{torch::dataloader},
#' containing the validation data.
#' @param device The device to perform validation on.
#' Default is 'cpu'; other options include 'gpu' and 'mps'.
#' @return A list containing the following elements:
#'   \describe{
#'     \item{accuracy_full_model}{Classification accuracy of the
#'     full (dense) model (if classification).}
#'     \item{accuracy_sparse}{Classification accuracy using only
#'     weights in active paths (if classification).}
#'     \item{validation_error}{Root mean squared error
#'     for the full model (if regression).}
#'     \item{validation_error_sparse}{Root mean squared error
#'     using only weights in active paths (if regression).}
#'     \item{density}{Proportion of weights with posterior
#'     inclusion probability > 0.5 in the whole network.}
#'     \item{density_active_path}{Proportion of weights
#'.    with inclusion probability > 0.5 after removing weights not in
#'.    active paths.}
#'   }
#' @export
validate_lbbnn <- function(LBBNN, num_samples, test_dl, device = "cpu") {
  LBBNN$eval()
  corrects <- 0
  corrects_sparse <- 0
  totals <- 0
  val_loss <- c()
  val_loss_mpm <- c()
  val_loss_mpm2 <- c()
  out_shape <- 1 #if binary classification or regression
  if (LBBNN$input_skip) {
    LBBNN$compute_paths_input_skip()
    }
  else (LBBNN$compute_paths)
  LBBNN$computed_paths <- TRUE
  torch::with_no_grad({
    coro::loop(for (b in test_dl){
      target <- b[[2]]$to(device = device)
      if (LBBNN$problem_type == "multiclass classification" | LBBNN$problem_type == "MNIST") { #nll loss needs float tensors but bce loss needs long tensors 
        target <- torch::torch_tensor(target, dtype = torch::torch_long())
        out_shape <- max(target)$item()
      }
      outputs <- torch::torch_zeros(num_samples, dim(b[[1]])[1],
                                    out_shape)$to(device = device)
      output_mpm <- torch::torch_zeros_like(outputs)
      for (i in 1:num_samples) {
        data <- b[[1]]$to(device = device)
        outputs[i] <- LBBNN(data, MPM = FALSE)
        output_mpm[i] <- LBBNN(data, MPM = TRUE)
      }
      out_full <- outputs$mean(1) #average over num_samples dimension
      out_mpm <- output_mpm$mean(1)
      if (LBBNN$problem_type == "multiclass classification" | LBBNN$problem_type == "MNIST") {
        prediction <- max.col(out_full)
        corrects <- corrects + sum(prediction == target)
        totals <- totals + length(target)
        #prediction using only weights in active paths
        prediction_mpm <- max.col(out_mpm)
        corrects_sparse <- corrects_sparse + sum(prediction_mpm == target)
      }
      else if (LBBNN$problem_type == "binary classification") {
        out_full <- out_full$squeeze()
        out_mpm <- out_mpm$squeeze()
        corrects <- corrects + sum((out_full > 0.5) == target)
        corrects_sparse <- corrects_sparse + sum((out_mpm > 0.5) == target)
        totals <- totals + length(target)
      }
      else {#for regression
        out_full <- out_full$squeeze()
        out_mpm <- out_mpm$squeeze()
        loss <- torch::torch_sqrt(torch::nnf_mse_loss(out_full, target))
        loss_mpm <- torch::torch_sqrt(torch::nnf_mse_loss(out_mpm, target))
        val_loss <- c(val_loss, loss$item())
        val_loss_mpm <- c(val_loss_mpm, loss_mpm$item())
      }
    })
  })
  acc_full <- corrects / totals
  acc_sparse <- corrects_sparse / totals
  density <- LBBNN$density()
  density2 <- LBBNN$density_active_path()
  if (LBBNN$problem_type != "regression") {
    l <- list("accuracy_full_model" = acc_full$item(), "accuracy_sparse" =
                acc_sparse$item(), "density" = density,
              "density_active_path" = density2)
  }
  else {
    l <- list("validation_error" = mean(val_loss),
              "validation_error_sparse" = mean(val_loss_mpm), "density" =
                density, "density_active_path" = density2)
  }
  return(l)
}

test_that("Smoke: tiny model trained for one epoch", {
  testthat::skip_on_cran()
  if (!requireNamespace("torch", quietly = TRUE)) {
    testthat::skip("torch not available")
  }
  i <- 5000
  j <- 15
  set.seed(2)
  torch::torch_manual_seed(2)
  #generate some data
  X <- matrix(rnorm(i * j, mean = - 0.1, sd = 0.1), ncol = j)
  #make some X relevant for prediction
  y_base <-  (0.1 * log(abs(X[, 1])) + 3 * cos(X[, 2]) + 2 * X[, 3] * X[, 4]
              + 2 * X[, 5] - 2 * X[, 6] ** 2 + rnorm(i, sd = 0.01))
  y <- c()
  # change y to 0 and 1
  y[y_base > median(y_base)] <- 1
  y[y_base <= median(y_base)] <- 0
  sim_data <- as.data.frame(X)
  sim_data <- cbind(sim_data, y)
  loaders <- get_dataloaders(sim_data, train_proportion = 0.9,
                             train_batch_size = 1500, test_batch_size = 500,
                             standardize = FALSE)
  train_loader <- loaders$train_loader
  test_loader  <- loaders$test_loader
  problem <- "binary classification"
  sizes <- c(j, 5, 5, 1) # 2 hidden layers, 5 neurons in each
  incl_priors <- c(0.5, 0.5, 0.5) #prior inclusion probs for each weight matrix
  stds <- c(100, 100, 100) #prior distribution for the standard deviation of the weights
  incl_inits <- matrix(rep(c(-15, 10), 3), nrow = 2, ncol = 3) #initializations for inclusion params
  device <- "cpu" #can also be 'gpu' or 'mps'
  model_input_skip <- lbbnn_net(problem_type = problem, sizes = sizes,
                                prior = incl_priors,
                                inclusion_inits = incl_inits, input_skip = TRUE,
                                std = stds, flow = TRUE, device = device)

  res <- train_lbbnn(epochs = 1, LBBNN = model_input_skip,
              lr = 0.005, train_dl = train_loader, device = device)
  validate_lbbnn(LBBNN = model_input_skip, num_samples = 1,
                 test_dl = test_loader, device)

  expect_true(length(res$loss) == 1)
  expect_true(is.numeric(res$density[1]))
})

#' @title Obtain adjacency matrices for \code{igraph} plotting
#' @description Takes the alpha active path matrices for each layer of the
#' LBBNN and converts them to adjacency matrices so that they can be plotted
#' with igraph.
#' @param model An instance of \code{lbbnn_net} with input_skip enabled.
#' @return A list of adjacency matrices, one for each hidden layer and output.
#' @keywords internal
get_adj_mats <- function(model) {
  mats_out <- list()
  i <- 1
  for (l in model$layers$children) {
    alp <- t(as.matrix(l$alpha_active_path))
    adj_mat <- matrix(0, nrow = sum(dim(alp)), ncol = sum(dim(alp))) #initialize empty matrix
    first_dim <- 1:dim(alp)[1]
    second_dim <- (dim(alp)[1] + 1):sum(dim(alp))
    adj_mat[first_dim, second_dim] <- alp
    mats_out[[i]] <- adj_mat
    i <- i + 1
  } #do the same for the output layer
  alp_out <- t(as.matrix(model$out_layer$alpha_active_path))
  adj_mat_out <- matrix(0, nrow = sum(dim(alp_out)), ncol = sum(dim(alp_out)))
  first_dim <- 1:dim(alp_out)[1]
  second_dim <- (dim(alp_out)[1] + 1):sum(dim(alp_out))
  adj_mat_out[first_dim, second_dim] <- alp_out
  mats_out[[i]] <- adj_mat_out
  return(mats_out)
}

#' @title Function for plotting nodes in the network between two layers.
#' @description Takes care of the three possible cases. Both layers have even
#' number of neurons, both layers have odd numbers, or one of each.
#' @param N integer, number of neurons in the first layer.
#' @param N_u integer, number of neurons in the second layer.
#' @param input_positions Positions of the neurons in the input layer.
#' @param neuron_spacing How much space between the neurons.
#' @return Positions of the second layer.
#' @keywords internal
assign_within_layer_pos <- function(N, N_u, input_positions,neuron_spacing) {
  if (N %% 2 == 0 && N_u %% 2 == 0) { #if both layers have even number of neurons
    N_u_center <- stats::median(input_positions)
    N_u_start_pos <- N_u_center + neuron_spacing / 2 - (N_u  / 2 * neuron_spacing) #add the half space, then subtract half of the array to get to start point
    N_u_positions <- seq(from = N_u_start_pos, length.out = N_u, by = neuron_spacing)
  }
  if (N %% 2 != 0 && N_u %% 2 != 0) { #if both layers have odd number of neurons
    N_u_center <- stats::median(input_positions)
    N_u_start_pos <- N_u_center - ((N_u - 1) / 2) * neuron_spacing #just need to figure out how many neurons to the left of the median one
    N_u_positions <- seq(from = N_u_start_pos, length.out = N_u, by = neuron_spacing)
  }
  if ((N + N_u) %% 2 != 0) { #in the case of even and odd number of neurons. Even + odd = odd
    if (N > N_u) { #in this case, N_u is odd
      N_u_center <- stats::median(input_positions)
      N_u_start_pos <- N_u_center + neuron_spacing / 2 - (N_u / 2 * neuron_spacing)
      N_u_positions <- seq(from = N_u_start_pos, length.out = N_u, by = neuron_spacing)
    }
    if (N < N_u) { #in this case, N_u is even
      N_u_center <- stats::median(input_positions)
      N_u_start_pos <- N_u_center - ((N_u - 1) / 2) * neuron_spacing
      N_u_positions <- seq(from = N_u_start_pos, length.out = N_u, by = neuron_spacing)
    }
  }
    return(N_u_positions)
}


#' @title Assign names to nodes.
#' @description Internal helper function to assign descriptive names to
#' nodes used for plotting.
#' @param model A trained \code{lbbnn_net} model with input-skip.
#' @return A list of adjacency matrices with the correct names.
#' @keywords internal
assign_names <- function(model) {#assign names to the nodes before plotting
  alphas <- get_adj_mats(model)
  sizes <- model$sizes
  for (i in 1:length(alphas)) {
    mat_names <- c()
    if (i == 1) { #for the input layer
      for (j in 1:sizes[1]) { #first the x_i
        name <- paste("x", j - 1, "_", i - 1, sep = "")
        mat_names <- c(mat_names, name)
      }
      for (j in 1:sizes[i + 1]) {#then the u
        name <- paste("u", j - 1, "_", i, sep = "")
        mat_names <- c(mat_names, name)
      }
      colnames(alphas[[i]]) <- mat_names
    }
    else if (i < length(alphas)) {#all other layers except the last
      mat_names <- c()
      for (j in 1:sizes[i]) {#N - n_input is the number of neurons in the hidden layer
        name <- paste("u", j - 1, "_", i - 1, sep = "")
        mat_names <- c(mat_names, name)
      }
      for (j in 1:sizes[1]) {#the input skip x
        name <- paste("x", j - 1, "_", i - 1, sep = "")
        mat_names <- c(mat_names, name)
      }
      for (j in 1:sizes[i + 1]) {#the hidden neurons for the next layer
        name <- paste("u", j - 1, "_", i, sep = "")
        mat_names <- c(mat_names, name)
      }
      colnames(alphas[[i]]) <- mat_names
    }
    else {#the last layer note: this is almost the same as above, could join them together??
      mat_names <- c()
      for (j in 1:sizes[i]){#N - n_input is the number of neurons in the hidden layer
        name <- paste("u", j - 1, "_", i - 1, sep = "")
        mat_names <- c(mat_names, name)
      }
      for (j in 1:sizes[1]) {#the input skip x
        name <- paste("x", j - 1, "_", i - 1, sep = "")
        mat_names <- c(mat_names, name)
      }
      for (j in 1:sizes[i + 1]){ #the hidden neurons for the next layer
        name <- paste("y", j - 1, sep = "")
        mat_names <- c(mat_names, name)
      }
      colnames(alphas[[i]]) <- mat_names
    }
  }
  return(alphas)
}

#' @title Function to plot an input skip structure after removing weights
#' in non-active paths.
#' @description Uses igraph to plot.
#' @param model A trained \code{lbbnn_net} model with input_skip enabled.
#' @param layer_spacing numeric, spacing in between layers.
#' @param neuron_spacing numeric, spacing between neurons within a layer.
#' @param vertex_size numeric, size of the neurons.
#' @param label_size numeric, size of the text within neurons.
#' @param edge_width numeric, width of the edges connecting neurons.
#' @param save_svg the path where the plot will be saved.
#' @examples
#' \donttest{
#' if (torch_available()) {
#' sizes <- c(2,3,3,2)
#' problem <- 'multiclass classification'
#' inclusion_priors <- c(0.1,0.1,0.1)
#' std_priors <- c(1.0,1.0,1.0)
#' inclusion_inits <- matrix(rep(c(-10,10),3), nrow = 2, ncol = 3)
#' device <- 'cpu'
#' torch::torch_manual_seed(0)
#' model <- lbbnn_net(problem_type = problem, sizes = sizes,
#'                  prior = inclusion_priors, inclusion_inits = inclusion_inits,
#'                   input_skip = TRUE, std = std_priors, flow = FALSE,
#'                   num_transforms = 2, dims = c(200,200), device = device)
#' model$compute_paths_input_skip()
#' LBBNN:::plot_active_paths(model, 1, 1, 14, 1)}
#' }
#' @return This function produces plots as a side effect and does not return a value.
plot_active_paths <- function(model, layer_spacing = 1, neuron_spacing = 1,
                              vertex_size = 10, label_size = 0.5,
                              edge_width = 0.5, save_svg = NULL) {

  if (!is.null(save_svg)) {
    # Open SVG device
    svglite::svglite(save_svg, width = 5, height = 4)
  }
  graph <- assign_names(model) #the graph with names neurons, given some model with alpha matrices
  g <- igraph::make_empty_graph(n = 0) #initialize empty graph
  for (L in 1:length(graph)) {
    g <- g +  igraph::graph_from_adjacency_matrix(graph[[L]], mode = "directed")
  }
  plot_points <- matrix(0, nrow = length(g), ncol = 2) #x,y coordinates for all neurons in g
  layer_positions <- seq(from = 0, by = - layer_spacing,
                         length.out = length(model$sizes)) #position for each layer
  index_start <- 0
  dim_1_pos <- 0
  i <- 1
  for (s in model$sizes) {
    if (i == 1) {
      plot_points[1:model$sizes[i], 2] <- layer_positions[i] #input layer coords
      index_start <- model$sizes[i] + 1 #where to start next layer
      dim_1_pos <- seq(from = 0, length.out = model$sizes[i], by = neuron_spacing)#coords within input layer
      plot_points[1:model$sizes[i], 1] <- dim_1_pos
    }
    else if (i < length(model$sizes)) {#all other layers except the last
      plot_points[(index_start:(index_start + model$sizes[1] + model$sizes[i] - 1)), 2] <- layer_positions[i]
      #N = size of prev layer #N_u size of current layer
      dim_1_pos <- assign_within_layer_pos(N = length(dim_1_pos),
                                         N_u = model$sizes[1] + model$sizes[i],
                                          input_positions = dim_1_pos,
                                          neuron_spacing = neuron_spacing)
      plot_points[(index_start:(index_start + model$sizes[1] + model$sizes[i] - 1)), 1] <- dim_1_pos
      index_start <- index_start + model$sizes[1] + model$sizes[i]
    }
    else { #output layer
      dim_1_pos <- assign_within_layer_pos(N = length(dim_1_pos),
                                        N_u = model$sizes[length(model$sizes)],
                                        input_positions = dim_1_pos
                                        , neuron_spacing = neuron_spacing)
      plot_points[(index_start:(dim(plot_points)[1])), 1] <- dim_1_pos
      plot_points[(index_start:(dim(plot_points)[1])), 2] <- layer_positions[i]
    }
    i <- i + 1
  }
  #assign colors based on what type of neuron it is
  for (z in 1:length(igraph::V(g))) {
    string <- igraph::V(g)[z]
    if (grepl("x", string$name)) { #for input neurons
      igraph::V(g)[z]$color <- "#D5E8D4"
      igraph::V(g)[z]$frame.color <- "#D5E8D4" #change color of boundary too
    }
    else if (grepl("u", string$name)) { #hidden neurons
      igraph::V(g)[z]$color <- "#ADD8E6"
      igraph::V(g)[z]$frame.color <- "#ADD8E6"
    }
    else {
      igraph::V(g)[z]$color <- "#F8CECC" #output neurons
      igraph::V(g)[z]$frame.color <- "#F8CECC"
    }
  }

  plot(g, vertex.size = vertex_size, vertex.label.cex = label_size,
       edge.color = "black", vertex.label.color = "black",
       edge.width = edge_width, layout = - plot_points[, 2:1],
       edge.arrow.mode = '-', margin = 0.0, asp = 0)
  if (!is.null(save_svg)) {
    message(paste("Plot saved as", save_svg))
    on.exit(grDevices::dev.off()) # ensures the device closes even if an error occurs
  }

}

#' @title Generate prior inclusion probabilities for the weights of a
#' LBBNN layer (linear or convolutional).
#' @description A function to generate prior probability values for
#' each weight in an LBBNN layer.
#' The same probability is applied to all weights within a layer.
#' @param x A number between 0 and 1.
#' @return a numeric to be added to either (out_shape,in_shape)
#' in case of linear layers, or (out_channels,in_channels,kernel0,kernel1)
#' in case of convolutional layers.
#' @keywords internal
alpha_prior <- function(x) {
  if (!is.numeric(x))
    stop("invalid_class:", " x must be numeric")
  if (x <= 0 || x >= 1)
    stop("invalid_value:", " x must be between 0 and 1")
  if (length(x) != 1)
    stop("only one value of x allowed per layer")
  return(x)
}

#' @title Generate prior standard deviation for weights and biases of either
#' linear or convolutional layers.
#' @description A function to generate prior standard deviations for
#' each weight and bias in an LBBNN layer.
#' @param x A number greater than 0.
#' @return a numeric to be added to either (out_shape,in_shape)
#' in case of linear layers or (out_channels,in_channels,kernel0,kernel1)
#' in case of convolutional layers.
#' @keywords internal
std_prior <- function(x) {
  if (!is.numeric(x))
    stop("invalid_class:", "x must be numeric")
  if (x <= 0)
    stop("invalid_value:", "x must be > 0")
  if (length(x) != 1)
    stop("only one value allowed per layer")
  return(x)
}

#' @title Generate initializations for the inclusion parameters.
#' @description Controls the initial density of the network, an important tuning
#' parameter. The initialization parameters are sampled from 1/(1+exp(-U)),
#' with U ~ Uniform(lower,upper).
#' @param lower numeric scalar.
#' @param upper numeric scalar, must be greater than
#' @return A numeric vector of length 2: \code{c(lower,upper)}
#' @keywords internal
density_initialization <- function(lower, upper) {
  if (!is.numeric(lower))
    stop("invalid_class:", " lower must be numeric")
  if (!is.numeric(upper))
    stop("invalid_class:", " upper must be numeric")
  if (length(lower) != 1)
    stop("lower bound can only have one value")
  if (length(upper) != 1)
    stop("upper bound can only have one value")
  if (lower >= upper)
    stop("the lower bound must be smaller than the upper bound")
  return(c(lower, upper))
}





#' @title Class to generate an LBBNN feed forward layer
#' @description This module implements a fully connected LBBNN layer.
#' It supports:
#' \itemize{
#'   \item Prior inclusion probabilities for weights and biases in each layer.
#'   \item Standard deviation priors for weights and biases in each layer.
#'   \item Optional normalizing flows (RNVP) for a more flexible posterior.
#'   \item Forward pass using either the full model,
#'    or the Median Probability Model (MPM).
#'   \item Computation of the KL-divergence.
#' }
#' @param in_features integer, number of input neurons.
#' @param out_features integer, number of output neurons.
#' @param prior_inclusion numeric scalar, prior inclusion probability
#' for each weight and bias in the layer.
#' @param standard_prior numeric scalar, prior standard deviation
#' for weights and biases in each layer.
#' @param density_init A numeric of size 2,
#' used to initialize the inclusion parameters, one for each layer.
#' @param flow logical, whether to use normalizing flows
#' @param num_transforms integer, number of transformations for \code{flow}.
#' Default is 2.
#' @param hidden_dims numeric vector, dimension of the hidden layer(s)
#' in the neural networks of the RNVP transform.
#' @param device The device to be used. Default is CPU.
#' @param bias_inclusion_prob logical, determines whether the bias
#' should be as associated with inclusion probabilities.
#' @param conv_net logical, whether the layer is used in a convolutional net.
#' @examples
#' \donttest{
#' if (torch_available()) {
#' l1 <- lbbnn_linear(in_features = 10,out_features = 5,prior_inclusion = 0.25,
#' standard_prior = 1,density_init = c(0,1),flow = FALSE)
#' x <- torch::torch_rand(20,10,requires_grad = FALSE)
#' output <- l1(x,MPM = FALSE) #the forward pass, output has shape (20,5)
#' print(l1$kl_div()$item())} #compute KL-divergence after the forward pass
#' }
#' @return A \code{torch::nn_module} object,
#' representing a fully connected LBBNN layer.
#' The module has the following methods:
#'   - \code{forward(input, MPM = FALSE)}: Computes activation
#'   (using the LRT at training time) of a batch of inputs.
#'   - \code{kl_div()}: Computes the KL-divergence.
#'   - \code{sample_z()}: Samples from the flow if \code{flow = TRUE},
#'   in addition returns the log-determinant of the Jacobian
#'   of the transformation.
#' @export
lbbnn_linear <- torch::nn_module(
  "lbbnn_linear",
  initialize = function(
    in_features,
    out_features,
    prior_inclusion,
    standard_prior,
    density_init,
    flow = FALSE,
    num_transforms = 2,
    hidden_dims = c(200, 200),
    device = "cpu",
    bias_inclusion_prob = FALSE,
    conv_net = FALSE) {
    self$in_features  <- in_features
    self$out_features <- out_features
    self$device <- device
    self$flow <- flow #true or false
    self$bias_inclusion_prob <- bias_inclusion_prob # true or false
    self$conv_net <- conv_net
    self$num_transforms <- num_transforms
    self$hidden_dims <- hidden_dims
    self$density_init <- density_initialization(density_init[1],
                                                density_init[2])
    #weight variational parameters

    self$weight_mean <- torch::nn_parameter(torch::torch_empty(out_features,
                                                               in_features,
                                                               device = device))
    self$weight_rho <- torch::nn_parameter(torch::torch_empty(out_features,
                                                              in_features,
                                                              device = device))
    self$weight_sigma <- torch::torch_empty(out_features, in_features,
                                            device = device)

    #bias variational parameters
    self$bias_mean <- torch::nn_parameter(torch::torch_empty(out_features,
                                                             device = device))
    self$bias_rho <- torch::nn_parameter(torch::torch_empty(out_features,
                                                            device = device))
    self$bias_sigma <- torch::torch_empty(out_features, device = device)

    #inclusion variational parameters
    self$lambda_l <- torch::nn_parameter(torch::torch_empty(out_features,
                                                            in_features,
                                                            device = device))
    self$alpha <- torch::torch_empty(out_features, in_features, device = device)
    self$alpha_active_path <- torch::torch_empty(out_features, in_features,
                                                 device = device)

    #define priors
    self$alpha_prior <- torch::torch_zeros(out_features, in_features, device =
                                             device) + alpha_prior(
                                             prior_inclusion)

    if (self$bias_inclusion_prob) {
    self$bias_lambda_l <- torch::nn_parameter(torch::torch_empty(out_features,
                                                                 device = device
                                                                 ))
    self$bias_alpha <- torch::torch_empty(out_features, device = device)

    #define priors for bias.
    self$bias_alpha_prior <- (torch::torch_zeros(out_features, device = device)
                              + alpha_prior(prior_inclusion))
    }

    #standard normal prior on the weights and biases
    self$weight_mean_prior <- torch::torch_zeros(out_features, in_features,
                                                 device = device)
    self$weight_sigma_prior <- (torch::torch_zeros(out_features, in_features,
                                                  device = device)
                                + std_prior(standard_prior))
    self$bias_mean_prior <- torch::torch_zeros(out_features, device = device)
    self$bias_sigma_prior <- (torch::torch_zeros(out_features, device = device)
                              + std_prior(standard_prior))

    #flow parameters
    if (self$flow) {
      self$q0_mean <- torch::nn_parameter(torch::torch_empty(in_features,
                                                             device = device))
      self$q0_logvar <- torch::nn_parameter(torch::torch_empty(in_features,
                                                               device = device))
      self$c1 <- torch::nn_parameter(torch::torch_empty(in_features,
                                                        device = device))
      self$b1 <- torch::nn_parameter(torch::torch_empty(in_features,
                                                        device = device))
      self$b2 <- torch::nn_parameter(torch::torch_empty(in_features,
                                                        device = device))

      self$RNVP_flow <- normalizing_flow(c(in_features, hidden_dims), "RNVP",
                                         self$num_transforms)
      self$RNVP_flow$to(device = device)
    }
    self$reset_parameters()
  },
  reset_parameters = function() {
    torch::nn_init_uniform_(self$weight_mean, -0.01, 0.01)
    torch::nn_init_normal_(self$weight_rho, mean = -9, std = 1.0)
    torch::nn_init_uniform_(self$bias_mean, -0.2, 0.2)
    torch::nn_init_normal_(self$bias_rho, mean = -9, std = 1.0)
    torch::nn_init_uniform_(self$lambda_l, self$density_init[1],
                            self$density_init[2])
    if (self$flow) {
      torch::nn_init_normal_(self$q0_mean, mean = 0, std = 1)
      torch::nn_init_normal_(self$q0_logvar, mean = -9, std = 1)
      torch::nn_init_normal_(self$c1, mean = 0, std = 1)
      torch::nn_init_normal_(self$b1, mean = 0, std = 1)
      torch::nn_init_normal_(self$b2, mean = 0, std = 1)
    }


  },
  sample_z = function() {
    q0_std <- torch::torch_sqrt(torch::torch_exp(self$q0_logvar))
    epsilon_z <- torch::torch_randn_like(q0_std)
    self$z <- self$q0_mean + q0_std * epsilon_z #the initial random var
    out <- self$RNVP_flow(self$z) #returns z_k and the log determinant
    return(out)
  }
  ,
  forward = function(input, MPM = FALSE) {
    self$alpha <- torch::torch_sigmoid(self$lambda_l)
    if (self$bias_inclusion_prob) {
    self$bias_alpha <- torch::torch_sigmoid(self$bias_lambda_l)
    }
    self$weight_sigma <- torch::torch_log1p(torch_exp(self$weight_rho))
    self$bias_sigma <- torch::torch_log1p(torch_exp(self$bias_rho))
    self$z_k <- torch::torch_ones_like(self$weight_mean) #if not flow, all z = 1
    if (self$flow) {
      out <- self$sample_z()
      self$z_k <- out$z
    }
    if (! MPM) {#compute  mean and the variance of the activations using the LRT
      e_w <- self$weight_mean * self$alpha * self$z_k
      var_w <- self$alpha * (self$weight_sigma^2 + (1 - self$alpha) *
                               self$weight_mean^2 * self$z_k^2)
      if (self$bias_inclusion_prob) {
        e_b_alpha <- self$bias_mean * self$bias_alpha
        var_b_alpha <- self$bias_alpha * (self$bias_sigma^2 +
                                      (1 - self$bias_alpha) * self$bias_mean^2)
      } else {
        e_b_alpha <- self$bias_mean
        var_b_alpha <- self$bias_sigma^2
      }
      e_b <- torch::torch_matmul(input, torch::torch_t(e_w)) + e_b_alpha
      var_b <- torch::torch_matmul(input^2, torch::torch_t(var_w)) + var_b_alpha
      eps <- torch::torch_randn(size = (dim(var_b)), device = self$device)
      activations <- e_b + torch::torch_sqrt(var_b) * eps

    }else {#median probability model
      w <- torch::torch_normal(self$weight_mean * self$z_k, self$weight_sigma)
      bias <- torch::torch_normal(self$bias_mean, self$bias_sigma)
      weight <- w * self$alpha_active_path

      #need the below if we are running a convolutional net without active paths
      if (self$conv_net) {
        gamma <- (self$alpha$clone()$detach() > 0.5) * 1.
        weight <- w * gamma
      }

      if (self$bias_inclusion_prob) {
      bias <- bias * (self$bias_alpha > 0.5)
      }
      activations <- torch::torch_matmul(input, torch::torch_t(weight)) + bias
    }

    return(activations)},
  kl_div = function() {
    z2 <- torch::torch_ones_like(self$weight_mean) #if not flow
    log_q <- 0 #these are 0 if no flow is used
    log_r <- 0
    if (self$flow) {
      out <- self$sample_z()
      z2 <- out$z
      log_det_q <- out$logdet
      log_q0 <- torch::torch_sum(-0.5 *
                                   torch::torch_log(pi)$to(device = self$device)
                                 - 0.5 * self$q0_logvar
                - 0.5 * ((self$z - self$q0_mean) ** 2
                         / torch::torch_exp(self$q0_logvar)))
      log_q <- log_q0 - log_det_q

      ######get log r
      W_mean <- z2 * self$weight_mean * self$alpha
      W_var <- self$alpha * (self$weight_sigma ^ 2 + (1 - self$alpha)
                             * self$weight_mean ^ 2 * z2 ^ 2)
      act_mean <- torch::torch_matmul(self$c1, torch::torch_t(W_mean))
      act_var <- torch::torch_matmul(self$c1^2, torch::torch_t(W_var))
      act_inner <- act_mean + torch::torch_sqrt(act_var) *
        torch::torch_randn_like(act_var)
      act <- torch::nnf_hardtanh(act_inner)
      mean_r <- self$b1$outer(act)$mean(-1)  # eq (9) from MNF paper
      log_var_r <- self$b2$outer(act)$mean(-1)  # eq (10) from MNF paper
      out2 <- self$sample_z()
      z_b <- out2$z
      log_det_r <- out2$logdet
      log_rb <- torch::torch_sum(-0.5 *
                                  torch::torch_log(pi)$to(device = self$device)
                                 - 0.5 * log_var_r
                                - 0.5 * ((z_b - mean_r) ** 2
                                         / torch::torch_exp(log_var_r)))
      log_r <- log_det_r + log_rb

    }

    if (self$bias_inclusion_prob) {
      kl_bias <- torch::torch_sum(
        self$bias_alpha * (
          torch::torch_log(self$bias_sigma_prior / self$bias_sigma)
          - 0.5
          + torch::torch_log(self$bias_alpha / self$bias_alpha_prior + 1e-20)
          + (self$bias_sigma^2 + (self$bias_mean - self$bias_mean_prior)^2)
          / (2 * self$bias_sigma_prior^2)
        )
        + (1 - self$bias_alpha) * torch::torch_log((1 - self$bias_alpha)
                                                 / (1 - self$bias_alpha_prior)
                                                 + 1e-20)
    )
    }else {
      kl_bias <- torch::torch_sum(
        torch::torch_log(self$bias_sigma_prior / self$bias_sigma)
        - 0.5
        + (self$bias_sigma^2 + (self$bias_mean - self$bias_mean_prior)^2) /
          (2 * self$bias_sigma_prior^2))
    }

    kl_weight <- torch::torch_sum(self$alpha * (torch::torch_log(
      self$weight_sigma_prior / self$weight_sigma) - 0.5 + torch::torch_log(
        self$alpha / self$alpha_prior + 1e-20) + (self$weight_sigma^2 +
        (self$weight_mean * z2 - self$weight_mean_prior)^2) / (
          2 * self$weight_sigma_prior^2)) + (1 - self$alpha) * torch::torch_log(
            (1 - self$alpha) / (1 - self$alpha_prior) + 1e-20))


    return(kl_bias + kl_weight + log_q - log_r)
    }
)

#' @title Class to generate an LBBNN convolutional layer.
#' @description
#'  It supports:
#' \itemize{
#'   \item Prior inclusion probabilities for weights and biases in each layer.
#'   \item Standard deviation priors for weights and biases in each layer.
#'   \item Optional normalizing flows (RNVP) for a more flexible posterior.
#'   \item Forward pass using either the full model or the
#'   Median Probability Model (MPM).
#'   \item Computation of the KL-divergence.
#' }
#' @param in_channels integer, number of input channels.
#' @param out_channels integer, number of output channels.
#' @param kernel_size size of the convolving kernel.
#' @param prior_inclusion numeric scalar, prior inclusion probability for
#' each weight and bias in the layer.
#' @param standard_prior numeric scalar, prior standard deviation for
#' weights and biases in each layer.
#' @param density_init A numeric of size 2, used to initialize the
#' inclusion parameters, one for each layer.
#' @param flow logical, whether to use normalizing flows
#' @param num_transforms integer, number of transformations for \code{flow}.
#'  Default is 2.
#' @param hidden_dims numeric vector, dimension of the hidden layer(s)
#' in the neural networks of the RNVP transform.
#' @param device The device to be used. Default is CPU.
#' @examples
#' \donttest{
#' if (torch_available()) {
#'layer <- lbbnn_conv2d(in_channels = 1,out_channels = 32,kernel_size = c(3,3),
#'prior_inclusion = 0.2,standard_prior = 1,density_init = c(0,1),device = 'cpu')
#'x <-torch::torch_randn(100,1,28,28)
#'out <-layer(x)
#'print(dim(out))}
#'}
#' @importFrom torch torch_empty torch_long torch_zeros torch_zeros_like
#' with_no_grad torch_rand torch_randn torch_exp
#' @return A \code{torch::nn_module} object representing a convolutional
#' LBBNN layer.
#' The module has the following methods:
#'   - \code{forward(input, MPM = FALSE)}: Computes activation
#'   (using the LRT at training time) of a batch of inputs.
#'   - \code{kl_div()}: Computes the KL-divergence.
#'   - \code{sample_z()}: Samples from the flow if \code{flow = TRUE},
#'   in addition returns the log-determinant of the Jacobian
#'  of the transformation.

#' @export
lbbnn_conv2d <- torch::nn_module(
  "lbbnn_conv2d",
  initialize = function(in_channels, out_channels, kernel_size, prior_inclusion,
                        standard_prior, density_init,
                        flow = FALSE, num_transforms = 2,
                        hidden_dims = c(200, 200), device = "cpu") {

    if (length(kernel_size) == 1) {
      kernel <- c(kernel_size, kernel_size)
    }
    else if (length(kernel_size) == 2) {
      kernel <- kernel_size
    }else (stop("kernel_size must be of either length one or two."))

    self$in_channels  <- in_channels
    self$out_channels <- out_channels
    self$flow <- flow #true or false
    self$num_transforms <- num_transforms
    self$hidden_dims <- hidden_dims
    self$device <- device
    self$density_init <- density_initialization(density_init[1],
                                                density_init[2])

    #weight variational parameters
    self$weight_mean <- torch::nn_parameter(torch_empty(out_channels,
                                                        in_channels, kernel[1],
                                                        kernel[2],
                                                        device = device))
    self$weight_rho <-  torch::nn_parameter(torch_empty(out_channels,
                                                        in_channels, kernel[1],
                                                        kernel[2],
                                                        device = device))
    self$weight_sigma <- torch::torch_empty(out_channels, in_channels,
                                            kernel[1], kernel[2],
                                            device = device)

    #bias variational parameters
    self$bias_mean <- torch::nn_parameter(torch_empty(out_channels,
                                                      device = device))
    self$bias_rho <- torch::nn_parameter(torch_empty(out_channels,
                                                     device = device))
    self$bias_sigma <- torch::torch_empty(out_channels, device = device)

    #inclusion variational parameters
    self$lambda_l <- torch::nn_parameter(torch_empty(out_channels, in_channels,
                                                     kernel[1], kernel[2],
                                                     device = device))
    self$alpha <- torch::torch_empty(out_channels, in_channels, kernel[1],
                                     kernel[2], device = device)

    #define priors.
    self$alpha_prior <- (torch::torch_zeros_like(
      self$alpha, device = self$device) + alpha_prior(prior_inclusion))

    #standard normal prior on the weights and biases
    self$weight_mean_prior <- torch::torch_zeros_like(self$weight_mean,
                                                      device = self$device)
    self$weight_sigma_prior <- (torch::torch_zeros_like(self$weight_sigma,
                                                        device = self$device)
                                + std_prior(standard_prior))
    self$bias_mean_prior <- torch::torch_zeros_like(self$bias_mean,
                                                    device = self$device)
    self$bias_sigma_prior <- (torch::torch_zeros_like(self$bias_sigma,
                                                      device = self$device)
                              + std_prior(standard_prior))

    #flow parameters
    if (self$flow) {
      self$q0_mean <- torch::nn_parameter(torch::torch_empty(out_channels,
                                                             device = device))
      self$q0_logvar <- torch::nn_parameter(torch::torch_empty(out_channels,
                                                               device = device))
      self$c1 <- torch::nn_parameter(torch::torch_empty(out_channels,
                                                        device = device))
      self$b1 <- torch::nn_parameter(torch::torch_empty(out_channels,
                                                        device = device))
      self$b2 <- torch::nn_parameter(torch::torch_empty(out_channels,
                                                        evice = device))


      self$RNVP_flow <- normalizing_flow(c(out_channels, hidden_dims), "RNVP",
                                         self$num_transforms)
      self$RNVP_flow$to(device = device)
    }

    self$reset_parameters()
  },
  reset_parameters = function() {
    torch::nn_init_uniform_(self$weight_mean, -0.2, 0.2)
    torch::nn_init_normal_(self$weight_rho, mean = -9, std = 0.1)
    torch::nn_init_uniform_(self$bias_mean, -0.2, 0.2)
    torch::nn_init_normal_(self$bias_rho, mean = -9, std = 0.1)
    torch::nn_init_uniform_(self$lambda_l, self$density_init[1],
                            self$density_init[2])
    if (self$flow) {
      torch::nn_init_normal_(self$q0_mean, mean = 0, std = 1)
      torch::nn_init_normal_(self$q0_logvar, mean = -9, std = 1)
      torch::nn_init_normal_(self$c1, mean = 0, std = 1)
      torch::nn_init_normal_(self$b1, mean = 0, std = 1)
      torch::nn_init_normal_(self$b2, mean = 0, std = 1)
    }

  },
  forward = function(input, MPM = FALSE) {
    self$alpha <- 1 / (1 + torch::torch_exp(- self$lambda_l))
    self$weight_sigma <- torch::torch_log1p(torch_exp(self$weight_rho))
    self$bias_sigma <- torch::torch_log1p(torch_exp(self$bias_rho))
    z_k <- torch::torch_ones(self$out_channels, device = self$device) #if MF
    if (self$flow) {
      out <- self$sample_z()
      z_k <- out$z
    }

    if (! MPM) {
      e_w <- self$weight_mean * self$alpha * z_k$view(c(-1, 1, 1, 1))
      var_w <- self$alpha * (self$weight_sigma ^ 2 + (1 - self$alpha) *
                            self$weight_mean ^ 2 * z_k$view(c(-1, 1, 1, 1)) ^ 2)
      psi <- torch::nnf_conv2d(input = input, weight = e_w, bias = self$bias_mu)
      delta <- torch::nnf_conv2d(input = input ^ 2, weight = var_w,
                                 bias = self$bias_sigma ^ 2)
      #delta[delta<= 0] = 0 +1e-20
      eps <- torch::torch_randn(size = (dim(delta)), device = self$device)
      activations <- psi + torch::torch_sqrt(delta) * eps
    }else {#only sample from mpm
      gamma <- (self$alpha$clone()$detach() > 0.5) * 1.
      w <- torch::torch_normal(self$weight_mean * z_k$view(c(-1, 1, 1, 1)),
                               self$weight_sigma)
      bias <- torch::torch_normal(self$bias_mean, self$bias_sigma)
      weight <- w * gamma
      activations <- torch::nnf_conv2d(input = input, weight = weight,
                                      bias = bias)
    }

    return(activations)},

  sample_z = function() {
    q0_std <- torch::torch_sqrt(torch::torch_exp(self$q0_logvar))
    epsilon_z <- torch::torch_randn_like(q0_std)
    self$z <- self$q0_mean + q0_std * epsilon_z
    out <- self$RNVP_flow(self$z)
    return(out)
  },

  kl_div = function() {
    z2 <- torch::torch_ones(self$out_channels, device = self$device)#if not flow
    log_q <- 0 #these are 0 if no flow is used
    log_r <- 0
    if (self$flow) {
      out <- self$sample_z()
      z2 <- out$z
      log_det_q <- out$logdet
      log_q0 <- torch::torch_sum(- 0.5 *
                                torch::torch_log(pi)$to(device = self$device)
                                - 0.5 * self$q0_logvar
                                - 0.5 * ((self$z - self$q0_mean) ** 2
                                / torch::torch_exp(self$q0_logvar)))
      log_q <- log_q0 - log_det_q

      ######get log r
      W_mean <- self$weight_mean * self$alpha * z2$view(c(-1, 1, 1, 1))
      W_var <-  self$alpha * (self$weight_sigma ^ 2 + (1 - self$alpha) *
                            self$weight_mean ^ 2 * z2$view(c(-1, 1, 1, 1)) ^ 2)
      act_mu <- torch::torch_matmul(W_mean$view(- 1, length(self$c1)), self$c1)
      act_var <- torch::torch_matmul(W_mean$view(- 1, length(self$c1)),
                                     self$c1 ^ 2)

      # For convolutional layers, linear mappings empirically work better than
      # tanh. Hence no need for act = tanh(act). Christos Louizos
      # confirmed this in https://github.com/AMLab-Amsterdam/MNF_VBNN/issues/4
      # even though the paper states the use of tanh in conv layers.
      act <- act_mu + torch::torch_sqrt(act_var) * torch::torch_randn_like(
        act_var)
      mean_r <- self$b1$outer(act)$mean(-1)  # eq (9) from MNF paper
      log_var_r <- self$b2$outer(act)$mean(-1)  # eq (10) from MNF paper
      out2 <- self$sample_z()
      z_b <- out2$z
      log_det_r <- out2$logdet
      log_rb <- torch::torch_sum(- 0.5 *
                                  torch::torch_log(pi)$to(device = self$device)
                                  - 0.5 * log_var_r
                                 - 0.5 * ((z_b - mean_r) ** 2
                                 / torch::torch_exp(log_var_r)))

      log_r <- log_det_r + log_rb

    }
    kl_bias <- torch::torch_sum(torch::torch_log(self$bias_sigma_prior
                                                 / self$bias_sigma)
                                - 0.5 + (self$bias_sigma ^ 2
                 + (self$bias_mean - self$bias_mean_prior) ^ 2)
                 / (2 * self$bias_sigma_prior ^ 2))

    kl_weight <- torch::torch_sum(self$alpha
                                  * (torch::torch_log(self$weight_sigma_prior
                                                      / self$weight_sigma)
                                  - 0.5 + torch::torch_log(self$alpha /
                                                      self$alpha_prior + 1e-20)
                                  + (self$weight_sigma ^ 2 +
                                       (self$weight_mean * z2$view(
                                         c(-1, 1, 1, 1))
                                        - self$weight_mean_prior) ^ 2) / (
                                        2 * self$weight_sigma_prior ^ 2))
                                  + (1 - self$alpha) * torch::torch_log(
                                    (1 - self$alpha) / (1 - self$alpha_prior)
                                    + 1e-20))

    return(kl_bias + kl_weight + log_q - log_r)
    }
)

#' Raisins Dataset
#'
#' @description Ilkay Cinar, Murat Kokl and Sakir Tasdemi(2020)
#' provide a dataset consisting of 2 varieties of Turkish raisins, with 450
#' samples of each type. The dataset contains 7 morphological features,
#' extracted from images taken of the Raisins. The goal is to classify to one of
#' the two types of Raisins.
#' @format this data frame has 900 rows and the following 8 columns:
#' \describe{
#'   \item{Area}{Number of pixels within the boundary}
#'   \item{MajorAxisLength}{Length of the main axis}
#'   \item{MinorAxisLength}{Length of the small axis}
#'   \item{Eccentricity}{Measure of the eccentricity of the ellipse}
#'   \item{ConvexArea}{The number of pixels of the
#'   smallest convex shell of the region formed by the raisin grain}
#'   \item{Extent}{Ratio of the region formed by the
#'   raisin grain to the total pixels in the bounding box}
#'   \item{Perimeter}{distance between the boundaries of the
#'   raisin grain and the pixels around it}
#'   \item{Class}{Kecimen or Besni raisin.}
#' }
#' @source \url{https://archive.ics.uci.edu/dataset/850/raisin}
"raisin_dataset"

#' Gallstone Dataset
#' @description Taken from the UCI machine learning repository. The task is
#' to classify whether the patient had gallstones or not. It contains a mix
#' of demographic data and bioimpedance data.
#' @format  This dataset has 319 rows and  38 columns.
#' @source \url{https://pmc.ncbi.nlm.nih.gov/articles/PMC11309733/#T2}
"gallstone_dataset"

library(torch)

#' @title Class to generate a normalizing flow
#' @param input_dim numeric vector, the dimensionality of each layer.
#' The first item is the input vector size.
#' @param transform_type Transformation type. Currently RNVP is implemented.
#' @param num_transforms integer, how many layers of transformations to include.
#' @description Used in\code{LBBNN_Net} when the argument \code{flow = TRUE}.
#' Contains a \code{torch::nn_module} where the initial vector gets transformed
#' through all the layers in the module.
#' Also computes the log-determinant of the Jacobian for the entire
#' transformation, the sum of the log-determinants of the independent layers.
#' @return
#' A \code{torch::nn_module} object representing the normalizing flow.
#' The module provides:
#' \describe{
#'   \item{\code{forward(z)}}{
#'     Applies all flow transformation layers to the input tensor \code{z}.
#'     Returns a named list containing:
#'     \describe{
#'       \item{\code{z}}{
#'         A `torch_tensor` containing the transformed version of the input,
#'         with the same shape as `z`.
#'       }
#'       \item{\code{logdet}}{
#'         A scalar `torch_tensor` equal to the sum of the log-determinants of
#'         all transformation layers.
#'       }
#'     }
#'   }
#' }
#' @examples
#' \donttest{
#'if (torch_available()) {
#'flow <- normalizing_flow(c(2,5,5), transform_type='RNVP', num_transforms = 3)
#'flow$to(device = 'cpu')
#'x <- torch::torch_rand(2, device = 'cpu')
#'output <- flow(x)
#'z_out <- output$z
#'print(dim(z_out))
#'log_det <- output$logdet
#'print(log_det)}
#' }
#' @export
normalizing_flow <- torch::nn_module(
  "normalizing_flow",
  initialize = function(input_dim, transform_type, num_transforms) {
    self$layers <- torch::nn_module_list()
    if (transform_type == "RNVP") {
      for (l in 1:num_transforms) {
        self$layers$append(rnvp_layer(input_dim))
      }
    }else {
      stop(paste("transform type", transform_type,
                 "not implemented, try 'RNVP' instead"))
    }
  },
  forward = function(z) {
    logdet <- 0
    for (l in self$layers$children) {
      z <- l(z)
      logdet <- logdet + l$log_det()
    }
    l = list("z" = z, "logdet" = logdet)
    return(l)
  }
)

#' Check if torch/libtorch is available for examples 
#' @export
torch_available <- function() {
  requireNamespace("torch", quietly = TRUE) &&
    torch::torch_is_installed() #without this, runtime operations will fail (missing libtorch)
}

#' @title Wrapper around \code{torch::dataloader}
#' @description  Avoids users having to manually define their own dataloaders.
#' @param dataset A \code{data.frame}.
#' The last column is assumed to be the dependent variable.
#' @param train_proportion numeric, between 0 and 1.
#' Proportion of data to be used for training.
#' @param train_batch_size integer, samples per batch in the train dataloader.
#' @param test_batch_size integer, samples per batch in the test dataloader.
#' @param standardize logical, standardize input-features, default is TRUE.
#' @param shuffle_train logical, shuffle training data each epoch. default TRUE
#' @param shuffle_test  logical, shuffle test data, default is FALSE.
#' @param seed integer. Used for reproducibility in the train/test split.
#' @return A list containing:
#'   \describe{
#'     \item{train_loader}{A \code{torch::dataloader} for the training data.}
#'     \item{test_loader}{A \code{torch::dataloader} for the test data.}
#'   }
#'@export
get_dataloaders <- function(dataset, train_proportion, train_batch_size,
                            test_batch_size, standardize = TRUE,
                            shuffle_train = TRUE, shuffle_test = FALSE,
                            seed = 1) {
  if (!inherits(dataset, "data.frame"))
    stop(paste("dataset must be a data.frame object"))
  if (!inherits(train_proportion, "numeric"))
    stop(paste(train_proportion, "must be a numeric value"))
  if (train_proportion > 1 || train_proportion < 0)
    stop(paste(train_proportion, "must be a numeric between 0 and 1"))
  set.seed(seed)
  sample <- sample.int(n = nrow(dataset),
                       size = floor(train_proportion * nrow(dataset)),
                       replace = FALSE)
  train  <- dataset[sample, ]
  test   <- dataset[-sample, ]
  p <- dim(train)[2] - 1
  y_train <- as.numeric(train[, dim(train)[2]])
  y_test <- as.numeric(test[, dim(test)[2]])
  x_train <- as.matrix(train[, 1:p])
  x_test <- as.matrix(test[, 1:p])
  if (min(y_train) == 0 && max(y_train) > 1) {
    y_train <- y_train + 1
    y_test <- y_test + 1
  }
  if (standardize) {
    x_train <- scale(x_train)
    x_test <- scale(x_test, center = attr(x_train, "scaled:center"),
                    scale = attr(x_train, "scaled:scale"))
  }
  train_data <- torch::tensor_dataset(torch::torch_tensor(x_train),
                                      torch::torch_tensor(y_train))
  test_data <- torch::tensor_dataset(torch::torch_tensor(x_test),
                                     torch::torch_tensor(y_test))
  if (train_batch_size > length(train_data))
    (stop("Can not have larger batch size than the amount of training data"))
  if (test_batch_size > length(test_data))
    (stop("Can not have larger test batch size than the amount of test data"))

  train_loader <- torch::dataloader(train_data, batch_size = train_batch_size,
                                    shuffle = shuffle_train)
  test_loader <- torch::dataloader(test_data, batch_size = test_batch_size,
                                   shuffle = shuffle_test)
  l <- list("train_loader" = train_loader, "test_loader" = test_loader)
  return(l)
}

library(Matrix)
require(graphics)

#' @title Function that checks how many times inputs are included,
#' and from which layer. Used in summary function.
#' @description Useful when the number of inputs and/or hidden neurons are very
#' large, and direct visualization of the network is difficult.
#' @param model An instance of \code{lbbnn_net} where \code{input_skip = TRUE}.
#' @return A matrix of shape (p, L-1) where p is the number of input variables
#' and L the total number of layers (including input and output),
#' with each element being 1 if the variable is included or 0 if not included.
#' @keywords internal
get_input_inclusions <- function(model) {
  if (model$input_skip == FALSE)
  (stop("This function is currently only implemented for input-skip"))
  x_names <- c()
  layer_names <- c()
  for (k in 1:model$sizes[1]) {
    x_names <- c(x_names, paste("x", k - 1, sep = ""))
  }
  for (l in 1:(length(model$sizes) - 1)) {
    layer_names <- c(layer_names, paste("L", l - 1, sep = ""))
  }
  inclusion_matrix <- matrix(0, nrow = model$sizes[1],
                             ncol = length(model$sizes) - 1)
  #add the names
  colnames(inclusion_matrix) <- layer_names
  rownames(inclusion_matrix) <- x_names
  inp_size <- model$sizes[1]
  i <- 1
  for (l in model$layers$children) {
    alp <- l$alpha_active_path
    incl <- as.numeric(torch::torch_sum(alp[, - inp_size:dim(alp)[2]], dim = 1))
    inclusion_matrix[, i] <- incl
    i <- i + 1
  }
  alp_out <- model$out_layer$alpha_active_path
  incl <- as.numeric(torch::torch_sum(alp_out[, - inp_size:dim(alp_out)[2]],
                                      dim = 1))
  inclusion_matrix[, i] <- incl
  i <- i + 1
  return(inclusion_matrix)
}

#' @title Summary of LBBNN fit
#' @description Summary method for objects of the \code{lbbnn_net} class.
#' Only applies to objects trained with \code{input_skip = TRUE}.
#' @param object An object of class \code{lbbnn_net}.
#' @param ... further arguments passed to or from other methods.
#' @details
#' The returned table combines two types of information:
#' \itemize{
#'   \item Number of times each input variable is included
#'         in the active paths from each layer
#'         (obtained from \code{get_input_inclusions()}).
#'   \item Average inclusion probabilities for each input variable from each layer,
#'.        including a final column showing the average across all layers.
#' }
#' @return A \code{data.frame} containing the above information.
#'   The function prints a formatted summary to the console.
#'   The returned \code{data.frame} is invisible.
#' @examples
#' \donttest{
#' if (torch_available()) {
#'x<-torch::torch_randn(3,2)
#'b <- torch::torch_rand(2)
#'y <- torch::torch_matmul(x,b)
#'train_data <- torch::tensor_dataset(x,y)
#'train_loader <- torch::dataloader(train_data,batch_size = 3,shuffle=FALSE)
#'problem<-'regression'
#'sizes <- c(2,1,1)
#'inclusion_priors <-c(0.9,0.2)
#'inclusion_inits <- matrix(rep(c(-10,10),2),nrow = 2,ncol = 2)
#'stds <- c(1.0,1.0)
#'model <- lbbnn_net(problem, sizes, inclusion_priors, stds, inclusion_inits,
#'flow = FALSE, input_skip = TRUE)
#'train_lbbnn(epochs = 1,LBBNN = model, lr = 0.01,train_dl = train_loader)
#'summary(model)}
#'}
#' @export
summary.lbbnn_net <- function(object, ...) {
  if (object$input_skip == FALSE)
  (stop("Summary only applies to objects with input-skip = TRUE"))
  if (object$computed_paths == FALSE) {
    object$compute_paths_input_skip()
    }
  inclusions <- get_input_inclusions(object) #a matrix of size (p,L), with p # inputs, L # layers

  #now get the average inclusion probabilities of each layer
  p <- object$sizes[1] # number of inputs
  L <- length(object$sizes) - 1 # number of layers
  alpha_means <- matrix(nrow = p, ncol = L)
  i <- 1
  all_alphas <- c()
  col_names <- c()
  for (l in object$layers$children) {
    alpha_l <- l$alpha$clone()$detach()$cpu()
    aa <- alpha_l[, (dim(alpha_l)[2] - p + 1):dim(alpha_l)[2]] #only need last p corresponding to input, ignoring hidden layer alphas
    alpha_means[, i] <- round(as.numeric(aa$mean(dim = 1)), 3)
    all_alphas <- rbind(all_alphas, as.matrix(aa))
    col_names <- c(col_names, paste("a", i - 1, sep = ""))
    i <- i + 1
  }
  #now do the output layer
  alpha_out <- object$out_layer$alpha$clone()$detach()
  a_out <- alpha_out[, (dim(alpha_out)[2] - p + 1):dim(alpha_out)[2]]
  all_alphas <- rbind(all_alphas, as.matrix(a_out))
  col_names <- c(col_names, paste("a", i - 1, sep = ""))
  alpha_means[, i] <- round(as.numeric(a_out$mean(dim = 1)), 3)
  a_avg <- round(colMeans(all_alphas), 3)
  colnames(alpha_means) <- col_names
  alpha_means <- cbind(alpha_means, a_avg)
  summary_out <- as.data.frame(cbind(inclusions, alpha_means))
  cat("Summary of lbbnn_net object:\n")
  cat("-----------------------------------\n")
  cat("Shows the number of times each variable was included from each layer\n")
  cat("-----------------------------------\n")
  cat("Then the average inclusion probability for each input from each layer\n")
  cat("-----------------------------------\n")
  cat("The final column shows the average inclusion probability across all layers\n")
  cat("-----------------------------------\n")
  print(summary_out)
  cat(paste("The model took", object$elapsed_time,
            "seconds to train, using", object$device))
  invisible(summary_out)

}

#' @title Residuals from LBBNN fit
#' @description Residuals from an object of the \code{lbbnn_net} class.
#' @param object An object of class \code{lbbnn_net}.
#' @param type Only 'response' is implemented i.e. y_true - y_predicted.
#' @param ... further arguments passed to or from other methods.
#' @return A numeric vector of residuals (\code{y_true - y_predicted})
#' @examples
#' \donttest{
#' if (torch_available()) {
#'x<-torch::torch_randn(3,2)
#'b <- torch::torch_rand(2)
#'y <- torch::torch_matmul(x,b)
#'train_data <- torch::tensor_dataset(x,y)
#'train_loader <- torch::dataloader(train_data,batch_size = 3,shuffle=FALSE)
#'problem<-'regression'
#'sizes <- c(2,1,1)
#'inclusion_priors <-c(0.9,0.2)
#'inclusion_inits <- matrix(rep(c(-10,10),2),nrow = 2,ncol = 2)
#'stds <- c(1.0,1.0)
#'model <- lbbnn_net(problem,sizes, inclusion_priors, stds ,inclusion_inits,
#'flow = FALSE, input_skip = TRUE)
#'train_lbbnn(epochs = 1,LBBNN = model, lr = 0.01,train_dl = train_loader)
#'residuals(model)}
#'}
#' @export
residuals.lbbnn_net <- function(object, type = c("response"), ...) {
  y_true <- object$y
  y_predicted <- object$r
  if (type == "response") {
    return(y_true - y_predicted)
  }
  else (stop("only y - y_pred residuals are currently implemented"))
}


#' @title Get model coefficients (local explanations) of \code{lbbnn_net} object
#' @description Given an input sample x_1,... x_j (with j the number of inputs,
#' the local explanation is found by considering active paths.
#' If relu activation functions are assumed, each path is a piecewise
#' linear function, so the contribution for x_j is just the sum of the weights
#' associated with the paths connecting x_j to the output.
#' The contributions are found by taking the gradient wrt x.
#' @param object an object of class \code{lbbnn_net}.
#' @param dataset Either a \code{torch::dataloader} object,
#' or a \code{torch::torch_tensor} object.
#' The former is assumed to be the same \code{torch::dataloader}
#' used for training or testing.
#' The latter can be any user-defined data.
#' @param inds Optional integer vector of row indices in the dataset to compute explanations for.
#' @param output_neuron integer, which output neuron to explain (default = 1).
#' @param num_data integer, if no indices are chosen,
#' the first \code{num_data} of \code{dataset} are automatically used for explanations.
#' @param num_samples integer, how many samples to use for model averaging
#' when sampling the weights in the active paths.
#' @param ... further arguments passed to or from other methods.
#' @details
#' \itemize{
#'   \item If \code{num_data = 1}, confidence intervals are computed using
#'          model averaging over \code{num_samples} weight samples.
#'   \item If \code{num_data > 1}, confidence intervals are computed across
#'.         the mean explanations for each sample.
#'   \item The output is a data frame with row names as input variables
#'         (\code{x0}, \code{x1}, \code{x2}, ...) and columns giving mean
#'         and 95% confidence intervals for each variable.
#' }
#' @return A data frame with rows corresponding to input variables
#' and the following columns:
#' \itemize{
#'   \item \code{lower}: lower bound of the 95% confidence interval
#'   \item \code{mean}: mean contribution of the variable
#'   \item \code{upper}: upper bound of the 95% confidence interval
#' }
#'
#' @examples
#' \donttest{
#' if (torch_available()) {
#'x<-torch::torch_randn(3,2)
#'b <- torch::torch_rand(2)
#'y <- torch::torch_matmul(x,b)
#'train_data <- torch::tensor_dataset(x,y)
#'train_loader <- torch::dataloader(train_data,batch_size = 3,shuffle=FALSE)
#'problem<-'regression'
#'sizes <- c(2,1,1)
#'inclusion_priors <-c(0.9,0.2)
#'inclusion_inits <- matrix(rep(c(-10,10),2),nrow = 2,ncol = 2)
#'stds <- c(1.0,1.0)
#'model <- lbbnn_net(problem,sizes,inclusion_priors,stds,inclusion_inits,flow = FALSE,
#'input_skip = TRUE)
#'train_lbbnn(epochs = 1,LBBNN = model, lr = 0.01,train_dl = train_loader)
#'coef(model,dataset = x, num_data = 1)}
#'}
#' @export
coef.lbbnn_net <- function(object, dataset, inds = NULL, output_neuron = 1,
                           num_data = 1, num_samples = 10, ...) {
  if (output_neuron > object$sizes[length(object$sizes)])
  stop (paste("output_neuron =", output_neuron, "can not be greater than",
              object$sizes[length(object$sizes)]))
  if (is.null(inds)) {
    all_means <- matrix(nrow = object$sizes[1], ncol = num_data)
  }
  else {
    all_means <- matrix(nrow = object$sizes[1], ncol = length(inds))
  }

  row_names <- c()
  for (i in 1:object$sizes[1]) {
    row_names <- c(row_names, paste("x", i - 1, sep = ""))
  }

  if (class(dataset)[1] == "dataloader") {
    X <- dataset$dataset$tensors[[1]]$clone()$detach()$cpu()
    }
  else if (class(dataset)[1] == "torch_tensor") {
  X <- dataset         # should be a tensor with shape (num_data,p)
  if (length(dim(X)) == 1) {
    X <- X$unsqueeze(dim = 1)
    } #reshape (p) to shape (1,p)
  if (dim(X)[length(dim(X))] != object$sizes[1])
  stop("the last index must have shape equal to p")
  }
  else {
    stop("dataset must be either a torch_tensor or a dataloader object")
    }

  if (is.null(inds)) {
    if (dim(X)[1] < num_data)
    stop(paste("num_data = ", num_data,
               "can not be greater than the number of total data points,", dim(X)[1]))
    X_explain <- X[1:num_data, ]

  }
  else {
    inds <- as.integer(inds) #in case user sends a numeric vector
    inds <- unique(inds) #remove any duplicates
    if (dim(X)[1] < length(inds))
    stop(paste("number of indecies =", length(inds),
      "can not be greater than the number of total data points,", dim(X)[1]))
    if (dim(X)[1] < max(inds))
    stop(paste("the largest index =", max(inds),
      "can not be greater than the number of total data points,", dim(X)[1]))
    num_data <- length(inds)
    X_explain <- X[inds, ]
  }

  if (num_data == 1) { #here we get the CI from the uncertainty around the one sample
    expl <- get_local_explanations_gradient(object, X_explain,
                                            num_samples = num_samples)
    e <- expl$explanations
    e <- e[, , output_neuron] #get the explanation based on which output
    mean_explanation <- as.matrix(e)
    qs <- t(apply(mean_explanation, 2, quants))
    colnames(qs) <- c("lower", "mean", "upper")
    rownames(qs) <- row_names
    return(as.data.frame(qs))
  }

  for (i in 1:num_data) { #loop over data points, here the CI is across the means for each sample
    data <- X_explain[i, ]
    expl <- get_local_explanations_gradient(object, data,
                                            num_samples = num_samples)
    e <- expl$explanations
    ee <- e[, , output_neuron]
    mean_explanation <- as.numeric(ee$mean(dim = 1))
    all_means[, i] <- mean_explanation
  }
  rownames(all_means) <- row_names
  qs <- t(apply(all_means, 1, quants))
  colnames(qs) <- c("lower", "mean", "upper")
  return(as.data.frame(qs))
}


#'@title Obtain predictions from the posterior of an \code{LBBNN model}
#'@description Draw from the posterior of a trained \code{lbbnn_net} object.
#'@param object A trained \code{lbbnn_net} object
#'@param mpm logical, whether to use the median probability model.
#'@param newdata A \code{torch::dataloader}
#'object containing the data with which to predict.
#'@param draws integer, the number of samples to draw from the posterior.
#'@param device character, device for computation (default = \code{"cpu"}).
#'@param link Optional link function to apply to the network output.
#'@param ... further arguments passed to or from other methods.
#'@return A \code{torch::torch_tensor}  of shape \code{(draws,N,C)}
#'where \code{N} is the number of samples in \code{newdata},
#'and \code{C} the number of outputs.
#' @examples
#' \donttest{
#' if (torch_available()) {
#'x<-torch::torch_randn(3,2)
#'b <- torch::torch_rand(2)
#'y <- torch::torch_matmul(x,b)
#'train_data <- torch::tensor_dataset(x,y)
#'train_loader <- torch::dataloader(train_data,batch_size = 3,shuffle=FALSE)
#'problem<-'regression'
#'sizes <- c(2,1,1)
#'inclusion_priors <-c(0.9,0.2)
#'inclusion_inits <- matrix(rep(c(-10,10),2),nrow = 2,ncol = 2)
#'stds <- c(1.0,1.0)
#'model <- lbbnn_net(problem,sizes,inclusion_priors,stds,inclusion_inits,
#'flow = FALSE,input_skip = TRUE)
#'train_lbbnn(epochs = 1,LBBNN = model, lr = 0.01,train_dl = train_loader)
#'predict(model,mpm = FALSE,newdata = train_loader,draws = 1)}
#'}
#' @export
predict.lbbnn_net <- function(object, newdata, mpm = FALSE, draws = 10,
                              device = "cpu", link = NULL, ...) {
  object$eval()
  object$raw_output <- TRUE #skip final sigmoid/softmax
  if (!object$computed_paths) {
    if (object$input_skip) {
      object$compute_paths_input_skip()
      }
    else (object$compute_paths)
  }
  if (class(newdata)[[1]] != "dataloader")
  stop("Currently only torch::dataloader objects are supported for newdata")
  out_shape <- object$sizes[length(object$sizes)] #number of output neurons
  all_outs <- NULL
  torch::with_no_grad({
    coro::loop(for (b in newdata){
      outputs <- torch::torch_zeros(draws, dim(b[[1]])[1],
                                    out_shape)$to(device = device)
      for (i in 1:draws) {
        data <- b[[1]]$to(device = device)
        outputs[i] <- object(data, MPM = mpm)
      }
      all_outs <- torch::torch_cat(c(all_outs, outputs), dim = 2) #add all the mini-batches together
    })
  })
  return(all_outs)
}

#' @title Print summary of an \code{lbbnn_net} object
#' @description
#' Provides a summary of a trained \code{lbbnn_net} object.
#' Includes the model type (input-skip or not), whether normalizing flows
#' are used, module and sub-module structure, number of trainable parameters,
#'  and prior variance and inclusion probabilities for the weights.
#' @param x An object of class \code{lbbnn_net}.
#' @param ... Further arguments passed to or from other methods.
#' @return Invisibly returns the input \code{x}.
#' @examples
#' \donttest{
#' if (torch_available()) {
#'x<-torch::torch_randn(3,2)
#'b <- torch::torch_rand(2)
#'y <- torch::torch_matmul(x,b)
#'train_data <- torch::tensor_dataset(x,y)
#'train_loader <- torch::dataloader(train_data,batch_size = 3,shuffle=FALSE)
#'problem<-'regression'
#'sizes <- c(2,1,1)
#'inclusion_priors <-c(0.9,0.2)
#'inclusion_inits <- matrix(rep(c(-10,10),2),nrow = 2,ncol = 2)
#'stds <- c(1.0,1.0)
#'model <- lbbnn_net(problem,sizes,inclusion_priors,stds,inclusion_inits,
#'flow = FALSE, input_skip = TRUE)
#'print(model)}
#'}
#' @export
print.lbbnn_net <- function(x, ...) {

  module_info <- x$modules[[1]]

  # Model description
  model_name <- if (isTRUE(x$input_skip)) {
    "LBBNN with input-skip"
  } else {
    "LBBNN without input-skip"
  }

  flow <- if (isTRUE(x$flow)) {
    "with normalizing flows"
  } else {
    "without normalizing flows"
  }

  # Header
  cat("\n========================================\n")
  cat("          LBBNN Model Summary           \n")
  cat("========================================\n\n")

  # Module info
  total_params <- sum(sapply(module_info$parameters, length))
  cat("Module Overview:\n")
  cat("  - An `nn_module` containing", total_params, "parameters.\n\n")

  # Submodules
  cat("---------------- Submodules ----------------\n")
  submodules <- module_info$modules
  if (length(submodules) == 0) {
    cat("  No submodules detected.\n")
  } else {
    for (name in names(submodules)) {
      mod <- submodules[[name]]
      if (is.null(mod)) next
      n_params <- if (!is.null(mod$parameters)) sum(sapply(mod$parameters, length)) else 0
      cat(sprintf("  - %-20s : %-15s # %d parameters\n", name, class(mod)[1], n_params))
    }
  }

  # Model details
  cat("\nModel Configuration:\n")
  cat("  -", model_name, "\n")
  cat("  - Optimized using variational inference", flow, "\n\n")

  # Priors
  cat("Priors:\n")
  cat("  - Prior inclusion probabilities per layer: ",
      paste(x$prior_inclusion, collapse = ", "), "\n")
  cat("  - Prior std dev for weights per layer:    ",
      paste(x$prior_std, collapse = ", "), "\n")

  cat("\n=================================================================\n\n")
  invisible(x)
}


#' @title Plot \code{lbbnn_net} objects
#' @description
#' Given a trained \code{lbbnn_net} model, this function produces either:
#' \itemize{
#'   \item \strong{Global plot}: a visualization of the network structure,
#'     showing only the active paths.
#'   \item \strong{Local explanation}: a plot of the local
#'     explanation for a single input sample, including error bars obtained
#'     from Monte Carlo sampling of the network weights.
#' }
#' @param x An instance of \code{lbbnn_net}.
#' @param type Either \code{"global"} or \code{"local"}.
#' @param data If local is chosen, one sample must be provided to obtain
#' the explanation. Must be a \code{torch::torch_tensor} of shape \code{(1,p)}.
#' @param num_samples integer, how many samples to use for model averaging
#' over the weights in case of local explanations.
#' @param ... further arguments passed to or from other methods.
#' @return No return value. Called for its side effects of producing a plot.
#' @export
plot.lbbnn_net <- function(x, type = c("global", "local"), data = NULL,
                           num_samples = 100, ...) {
  if (x$input_skip == FALSE)
  (stop("Plotting currently only implemented for input-skip"))
  if (x$computed_paths == FALSE)  {
    x$compute_paths_input_skip()
    }
  d <- match.arg(type)
  if (d == "global") {
    plot_active_paths(x, ...)
  }
  else {
    if (is.null(data))
    stop("data must contain a sample to explain")
    plot_local_explanations_gradient(x, input_data = data,
                                     num_samples = num_samples,
                                     device = x$device, ...)
  }

}

#' @title Get gradient based local explanations for input-skip LBBNNs.
#' @description Works by computing the gradient wrt to input, given we have
#' relu activation functions.
#' @param model A \code{LBBNN_Net} with input-skip
#' @param input_data The data to be explained (one sample).
#' @param num_samples integer, samples to use to produce credible intervals.
#' @param magnitude If TRUE, only return explanations.
#' If FALSE, multiply by input values.
#' @param include_potential_contribution IF TRUE, If covariate=0,
#' we assume that the contribution is negative
#' (good/bad that it is not included) if FALSE, just removes zero covariates.
#' @param device character, the device to be trained on.
#' Default is 'cpu', can be 'mps' or 'gpu'.
#' @return A list with the following elements:
#'   \describe{
#'     \item{explanations}{A \code{torch::tensor}
#'     of shape (num_samples, p, num_classes).}
#'     \item{p}{integer, the number of input features.}
#'     \item{predictions}{A \code{torch::tensor}
#'     of shape (num_samples, num_classes).}
#'   }
#' @export
get_local_explanations_gradient <- function(model, input_data, num_samples = 1,
                                            magnitude = TRUE,
                                        include_potential_contribution = FALSE,
                                        device = "cpu") {
  if (model$input_skip == FALSE) (stop("This function is only implemented for input-skip"))
  if (model$computed_paths == FALSE) {
    model$compute_paths_input_skip() #need this before computing explanations
  }
  #need to make sure input_data comes in shape (1,p) where p is #input variables
  num_classes <- model$sizes[length(model$sizes)]
  if (input_data$dim() == 4) { #in the case of MNIST or other image data
    (input_data <- input_data$view(c(-1, dim(input_data)[3] * dim(input_data)[4]))) #reshape to(1,p)

  }
  else { #in case shape is either (p) or (1,p), both cases returns (1,p)
    input_data <- input_data$view((c(-1, length(input_data))))
  }
  p <- input_data$shape[2] #number of variables
  explanations <- torch::torch_zeros(num_samples, p, num_classes)
  predictions <- torch::torch_zeros(num_samples, num_classes)
  model$raw_output <- TRUE #to skip last sigmoid/softmax layer
  for (b in 1:num_samples) {#for each sample, get explanations for each class
    input_data$requires_grad <- TRUE
    input_data <- input_data$to(device = device)
    model$zero_grad()
    output <- model(input_data, MPM = TRUE) #forward pass, using MPM
    for (k in 1:num_classes) {
      output_value <- output[1, k]
      grad <- torch::autograd_grad(outputs = output_value, inputs = input_data,
                            grad_outputs = torch::torch_ones_like(output_value),
                                  retain_graph = TRUE)
      explanations[b, , k] <- grad[[1]]
      predictions[b, k] <- output[1, k]
    }

  }

  inds <- torch::torch_nonzero(input_data == 0)[, 2] #find index of 0s
  if (include_potential_contribution) {
      explanations[, inds] <- - explanations[, inds]
  }
  else {#remove variables that do not contribute to predictions
    explanations[, inds] <- 0
  }
  if (! magnitude) {
    explanations <- explanations * input_data$view(c(1, -1, 1))
  }
  predictions <- predictions$detach()$cpu()
  outputs <- list("explanations" = explanations, "p" = p, "predictions" = predictions)
  return(outputs)
}

#' @title Function to obtain empirical 95% confidence interval, including the mean
#' @description Using the built in quantile function to return 95% confidence interval
#' @param x numeric vector whose sample quantiles is desired.
#' @return The quantiles in addition to the mean.
#' @export
quants <- function(x) { #maybe should allow for something other than 95% CI
  out <- c(
    lower = stats::quantile(x, 0.025),
    mean  = mean(x),
    upper = stats::quantile(x, 0.975)
  )
  return(out)
}


#' @title Plot the gradient based local explanations for one sample.
#' @description Plots the contribution of each covariate,
#' and the prediction, with error bars.
#' @param model An instance of \code{LBBNN_Net} with input-skip enabled.
#' @param input_data The data to be explained (one sample).
#' @param num_samples integer, samples to use to produce credible intervals.
#' @param device character, the device to be trained on. Default is cpu.
#' Can be 'mps' or 'gpu'.
#' @param save_svg the path where the plot will be saved as svg,
#' if save_svg is not NULL.
#' @return This function produces plots as a side
#' effect and does not return a value.
plot_local_explanations_gradient <- function(model, input_data, num_samples,
                                             device = "cpu", save_svg = NULL) {
  outputs <- get_local_explanations_gradient(model = model,
                                             input_data = input_data,
                                             num_samples = num_samples,
                                             device = device)

  preds<- as.matrix(outputs$predictions) #shape (num_samples,num_classes)
  expl <- as.array(outputs$explanations) #shape (num_samples,p,num_classes)
  for (cls in 1:model$sizes[length(model$sizes)]) {

    expl_class <- expl[, , cls]
    expl_quantiles <- apply(expl_class, 2, quants)
    pred_quantiles <- apply(preds, 2, quants)
    names <- c()
    median <- expl_quantiles[2, ]
    pred_median <- pred_quantiles[, cls][2] #add the median of the prediction
    contribution <- c(median, pred_median)
    min <- c(expl_quantiles[1, ], pred_quantiles[, cls][1])
    max <- c(expl_quantiles[3, ], pred_quantiles[, cls][3])
    for (x in 1:model$sizes[1]) {#get names for x-axis
      name <- paste("x", x - 1, sep = "")
      names <- c(names, name)

    }
    output <- "out"
    names <- c(names, output)
    data <- data.frame(
      name = names,
      contribution = contribution,
      min = min,
      max = max
    )


    plt <- ggplot2::ggplot(data <- data, ggplot2::aes(x = factor(name, levels = name),
                            y = contribution,
             fill = factor(ifelse(name == "out", "out", "input variables")))) +
      ggplot2::geom_bar(stat = "identity") +
      ggplot2::scale_fill_manual(name = paste("Output neuron", cls), values = c("#D5E8D4", "#F8CECC")) +
      ggplot2::geom_errorbar(ggplot2::aes(x = name, ymin = min, ymax = max),
                             width = 0.6, colour = "black", alpha = 0.9, size = 0.5) +
      ggplot2::xlab("") + ggplot2::ylab("Contribution") + ggplot2::ggtitle(
        "Local explanation, with 95% empirical confidence bars")
    print(plt)
    if (!is.null(save_svg)) {
      new_file <- file.path(
        dirname(save_svg),
        paste0(sub("\\.[^.]*$", "", basename(save_svg)), "_class_", cls, ".svg") #check how this works
      )
      svglite::svglite(new_file, width = 8, height = 6)
      print(plt)
      grDevices::dev.off()
      message("SVG saved to: ", new_file)
    }

  }

}

library(testthat)
library(torch)
library(torchvision)
library(LBBNN)

test_that("KMNIST batch has correct shape", {
  

  
  batch_size <- 100
  dir <- "./dataset/kmnist"
  
  # Transform function to ensure channel dimension exists
  kmnist_transform <- function(x) {
    print("before transform:")
    print(dim(x))
    d <- dim(x)
    if (length(d) == 3 && d[3] > 1 && d[1] == d[2]) {#if shape [28,28,batch] as on windows and linux(?)
      x <- torchvision::transform_to_tensor(x) #now shape should be [batch, 28,28]
      x <- x$unsqueeze(2) #add the channel dimension - > [batch,1,28,28]
    }
    else{ #on mac, everything is fine and easy
      x <- torchvision::transform_to_tensor(x)
    }
    print("after transform:")
    print(dim(x))
    x
  }
  
  # Load datasets
  train_ds <- torchvision::kmnist_dataset(
    dir,
    download = TRUE,
    transform = kmnist_transform
  )
  
  train_loader <- torch::dataloader(train_ds, batch_size = batch_size, shuffle = TRUE)
  
  # Grab first batch
  it <- train_loader$.iter()
  batch <- it$.next()[[1]]
  
  print("Batch shape:")
  print(dim(batch))
  
  # Check that batch has shape [batch_size, 1, 28, 28]
  expect_equal(dim(batch)[1], batch_size)
  expect_equal(dim(batch)[2], 1)
  expect_equal(dim(batch)[3], 28)
  expect_equal(dim(batch)[4], 28)
})

library(testthat)
library(torch)
library(torchvision)
library(coro)
library(LBBNN)

test_that("KMNIST batch has correct shape", {
  
  batch_size <- 100
  dir <- "./dataset/kmnist"
  
  kmnist_transform <- function(x) {
    print("before transform:")
    print(dim(x))
    d <- dim(x)
    if (length(d) == 3 && d[3] > 1 && d[1] == d[2]) {
      x <- torchvision::transform_to_tensor(x)
      x <- x$unsqueeze(2)
    } else {
      x <- torchvision::transform_to_tensor(x)
    }
    print("after transform:")
    print(dim(x))
    x
  }
  
  train_ds <- torchvision::kmnist_dataset(
    dir,
    download = TRUE,
    transform = kmnist_transform
  )
  
  train_loader <- torch::dataloader(train_ds, batch_size = batch_size, shuffle = TRUE)
  
  it <- train_loader$.iter()
  batch <- it$.next()[[1]]
  
  print("Batch shape:")
  print(dim(batch))
  
  expect_equal(dim(batch)[1], batch_size)
  expect_equal(dim(batch)[2], 1)
  expect_equal(dim(batch)[3], 28)
  expect_equal(dim(batch)[4], 28)
})


# ---------------------------------------------------------
# 🔴 NEW TEST: Reproduce prediction accumulation issue
# ---------------------------------------------------------

test_that("LBBNN ConvNet prediction accumulation is stable", {
  
  skip_on_cran()  # avoid CRAN timeouts
  
  torch::torch_manual_seed(1)
  
  batch_size <- 100
  draws <- 3
  out_dim <- 10
  dir <- "./dataset/kmnist"
  device <- "cpu"
  
  kmnist_transform <- function(x) {
    x <- torchvision::transform_to_tensor(x)
    if (length(dim(x)) == 3) {
      x <- x$unsqueeze(1)
    }
    x
  }
  
  test_ds <- torchvision::kmnist_dataset(
    dir,
    train = FALSE,
    download = TRUE,
    transform = kmnist_transform
  )
  
  test_loader <- torch::dataloader(test_ds, batch_size = batch_size)
  
  # --- YOUR MODEL (no training) ---
  
  conv_layer_1 <- lbbnn_conv2d(1, 32, 5, 0.5, 1, c(-10, 10), 2, FALSE, c(200,200), device)
  conv_layer_2 <- lbbnn_conv2d(32, 64, 5, 0.5, 1, c(-10, 15), 2, FALSE, c(200,200), device)
  
  linear_layer_1 <- lbbnn_linear(1024, 300, 0.5, 1, c(-10, 10), 2, FALSE, c(200,200), device,
                                 bias_inclusion_prob = FALSE, conv_net = TRUE)
  
  linear_layer_2 <- lbbnn_linear(300, 10, 0.5, 1, c(-5, 15), 2, FALSE, c(200,200), device,
                                 bias_inclusion_prob = FALSE, conv_net = TRUE)
  
  model <- LBBNN_ConvNet(conv_layer_1, conv_layer_2,
                         linear_layer_1, linear_layer_2, device)
  
  model$eval()
  
  # --- ORIGINAL BUG-PRONE LOGIC ---
  
  predictions <- NULL
  
  torch::with_no_grad({ 
    coro::loop(for (b in test_loader) { 
      
      outputs <- torch::torch_zeros(draws, dim(b[[1]])[1], out_dim)
      
      for (i in 1:draws) {
        data <- b[[1]]
        outputs[i] <- model(data, MPM = TRUE, predict = TRUE)
      }
      
      predictions <- torch::torch_cat(c(predictions, outputs), dim = 2)
    })  
  })
  
  print("Final prediction shape:")
  print(dim(predictions))
  
  # --- ASSERTIONS ---
  
  expect_true(length(dim(predictions)) == 3)
  expect_equal(dim(predictions)[1], draws)
  expect_equal(dim(predictions)[3], out_dim)
  expect_equal(dim(predictions)[2], length(test_ds))  # should be 10000
  
})

---
title: "Getting started with LBBNN"
output:
  rmarkdown::html_vignette:
    df_print: paged
params:
  eval: false
vignette: >
  %\VignetteIndexEntry{Getting started with LBBNN}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = params$eval
)
```

## Introduction

LBBNN implements Latent Bayesian Binary Neural Networks in R using the torch framework. This vignette walks through basic usage: data preparation, model definition, training, validation, and visualization.

## Setup

```{r}
library(LBBNN)
library(ggplot2)
library(torch)
```

## Data loaders

```{r}
loaders <- get_dataloaders(Raisin_Dataset, train_proportion = 0.8,
                           train_batch_size = 720, test_batch_size = 180)
train_loader <- loaders$train_loader
test_loader  <- loaders$test_loader
```

## Define the model

```{r}
problem <- "binary classification"
sizes <- c(7, 5, 5, 1)
inclusion_priors <- c(0.5, 0.5, 0.5)
stds <- c(1, 1, 1)
inclusion_inits <- matrix(rep(c(-10, 15), 3), nrow = 2, ncol = 3)
device <- "cpu"
torch_manual_seed(0)
model_input_skip <- lbbnn_net(problem_type = problem, sizes = sizes,
                              prior = inclusion_priors,
                              inclusion_inits = inclusion_inits,
                              input_skip = TRUE, std = stds,
                              flow = FALSE, device = device)
```

## Train

```{r}
results_input_skip <- train_lbbnn(epochs = 50, LBBNN = model_input_skip,
                                  lr = 0.005, train_dl = train_loader,
                                  device = device)
```

## Validate

```{r}
validate_lbbnn(LBBNN = model_input_skip, num_samples = 100,
               test_dl = test_loader, device = device)
```

---
title: "variable selection: non-linear data"
output: rmarkdown::html_vignette
params:
  eval: true
vignette: >
  %\VignetteIndexEntry{variable selection: non-linear data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(LBBNN)
has_torch <- requireNamespace("torch", quietly = TRUE) &&
            torch::torch_is_installed()
```

## Generate data

We again generate 1000 samples with 15 features, and make 6 of them relevant for the outcome. This is a more complicated problem
than the previous, as we include some non-linear effects. This time we transform the outcome into a binary variable. 
```{r, eval = has_torch}
i <- 1000
j <- 15
set.seed(42)
torch::torch_manual_seed(42)
X_nl <- matrix(runif(i * j, 0, 0.5), ncol = j)
y_nl <- (- 3 +  0.1 * log(abs(X_nl[, 1])) + 3 * cos(X_nl[, 2]) 
             + 2 * X_nl[, 3] * X_nl[, 4] + X_nl[, 5] - 
               X_nl[, 6] ** 2 + rnorm(i, sd = 0.1))
y <- c()
# change y to 0 and 1
y[y_nl > median(y_nl)] <- 1
y[y_nl <= median(y_nl)] <- 0
sim_data_nl <- as.data.frame(X_nl)
sim_data_nl <- cbind(sim_data_nl, y)
loaders_nl <- get_dataloaders(sim_data_nl, train_proportion = 0.9,
                           train_batch_size = 450, test_batch_size = 100,
                           standardize = FALSE)
train_loader_nl <- loaders_nl$train_loader
test_loader_nl  <- loaders_nl$test_loader
```

## Define hyperparameters and the model object

We use the same architecture as in the example with linear data. For this example, 
we use normalizing flows in the variational distribution. 
```{r, eval = has_torch}
problem <- "binary classification"
sizes <- c(j, 5, 5, 1) 
incl_priors <- c(0.5, 0.5, 0.5) 
stds <- c(1, 1, 1) 
incl_inits <- 'polarized'
device <- "cpu" 
model_nl <- lbbnn_net(problem_type = problem, sizes = sizes,
                              prior = incl_priors,
                              inclusion_inits = incl_inits, input_skip = TRUE,
                              std = stds, flow = TRUE, dims = c(10, 10, 10),
                              device = device, bias_inclusion_prob = FALSE)
```

## Train and validate the model

```{r, eval = has_torch}
train_lbbnn(epochs = 20, LBBNN = model_nl,
            lr = 0.2, train_dl = train_loader_nl, device = device, verbose = FALSE)

validate_lbbnn(LBBNN = model_nl, num_samples = 2, test_dl = test_loader_nl,
               device = device)
```

## Check the global explanations

```{r,,fig.width=6, fig.height=6, eval = has_torch}
plot(model_nl, type = "global", vertex_size = 7,
     edge_width = 0.4, label_size = 0.4)
```

All the relevant features are included. 

---
title: "classification: convolutional architecture"
output: rmarkdown::html_vignette
params:
  eval: true
vignette: >
  %\VignetteIndexEntry{classification: convolutional architecture}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(LBBNN)
has_torch <- requireNamespace("torch", quietly = TRUE) &&
            torch::torch_is_installed()
```
The goal of this vignette is to demonstrate how to use LBBNNs with a convolutional architecture.
We will only use a dummy dataset here, but the torchvision package can be used to try it out
on real datasets e.g. MNIST. 

## Demonstration of how to download KMNIST using torchvision

```{r,eval=FALSE}
if(!requireNamespace("torchvision"))
  install.packages("torchvision")
torch::torch_manual_seed(42)
dir <- "./dataset/kmnist"
kmnist_transform <- function(x) {
  d <- dim(x)
  if (length(d) == 3 && d[3] > 1 && d[1] == d[2]) {#if shape [28,28,batch] as on windows and linux(?)
    x <- torchvision::transform_to_tensor(x) #now shape should be [batch, 28,28]
    x <- x$unsqueeze(2) #add the channel dimension - > [batch,1,28,28]
  }
  else{ #on mac, everything is fine 
    x <- torchvision::transform_to_tensor(x)
  }
  return(x)
}
#get datasets from torchvision and define training and test loaders
train_ds <- torchvision::kmnist_dataset(
  dir,
  download = TRUE,
  transform = kmnist_transform)

test_ds <- torchvision::kmnist_dataset(
  dir,
  train = FALSE,
  transform = kmnist_transform)

train_loader_kmnist <- torch::dataloader(train_ds, batch_size = 100, shuffle = TRUE)
test_loader_kmnist <- torch::dataloader(test_ds, batch_size = 100)
```


## Create dummy dataset with the same shape as KMNIST

```{r, eval = has_torch}
torch::torch_manual_seed(42)
x <- torch::torch_randn(200, 1, 28, 28)
y <- torch::torch_randint(1, 11, size = 200)
dataset <- torch::tensor_dataset(x, y)
train_loader <- torch::dataloader(dataset, batch_size = 100)
```

## Create the layers that define the architecture of our convolutional network

Here we manually define the layers, as we will use both convolutional and feed-forward layers.
They are defined as follows:
```{r, eval = has_torch}
device <- "cpu"
conv_layer_1 <- lbbnn_conv2d(in_channels = 1, out_channels = 32, kernel_size = 5,
                             prior_inclusion = 0.5, standard_prior = 1,
                             density_init = c(-10, 10), num_transforms = 2,
                             flow = FALSE, hidden_dims = c(200, 200),
                             device = device)
conv_layer_2 <- lbbnn_conv2d(in_channels = 32, out_channels = 64, kernel_size = 5,
                             prior_inclusion = 0.5, standard_prior = 1,
                             density_init = c(-10, 15), num_transforms = 2,
                             flow = FALSE, hidden_dims = c(200, 200),
                             device = device)

linear_layer_1 <- lbbnn_linear(in_features = 1024, out_features = 300,
                               prior_inclusion = 0.5, standard_prior = 1,
                               density_init = c(-10, 10), num_transforms = 2,
                               flow = FALSE, hidden_dims = c(200, 200), device = device,
                               bias_inclusion_prob = FALSE, conv_net = TRUE)

linear_layer_2 <- lbbnn_linear(in_features = 300, out_features = 10,
                               prior_inclusion = 0.5, standard_prior = 1,
                               density_init = c(-5, 15),num_transforms = 2,
                               flow = FALSE, hidden_dims = c(200, 200), device = device,
                               bias_inclusion_prob = FALSE, conv_net = TRUE)
```

## Define the model object

We include pooling layers between the convolutional layers.

```{r, eval = has_torch}
LBBNN_ConvNet <- torch::nn_module(
  "LBBNN_ConvNet",
  
  initialize = function(conv1, conv2, fc1 ,fc2 ,device = device) {
    self$problem_type <- "multiclass classification"
    self$input_skip <- FALSE
    self$conv1 <- conv1
    self$conv2 <- conv2
    self$fc1 <- fc1
    self$fc2 <- fc2
    self$pool <- torch::nn_max_pool2d(2)
    self$act <- torch::nn_leaky_relu()
    self$out <- torch::nn_log_softmax(dim = 2)
    self$pout <- torch::nn_softmax(dim = 2)
    self$loss_fn <- torch::nn_nll_loss(reduction = "sum")
  },
  
  forward = function(x, MPM = FALSE, predict = FALSE) {
    x = self$act(self$conv1(x, MPM))
    x = self$pool(x)
    x = self$act(self$conv2(x, MPM))
    x = self$pool(x)
    x = torch::torch_flatten(x,start_dim = 2)
    x = self$act(self$fc1(x, MPM))
    if(!predict)
      x = self$out(self$fc2(x ,MPM))
    else
      x = self$pout(self$fc2(x ,MPM))
  },
  kl_div = function(){
    kl <- self$conv1$kl_div() + self$conv2$kl_div() +
      self$fc1$kl_div() + self$fc2$kl_div()
    return(kl)
  },
  density = function(){
    alphas <- NULL
    alphas <- c(as.numeric(self$conv1$alpha), as.numeric(self$conv2$alpha)
                ,as.numeric(self$fc1$alpha), as.numeric(self$fc2$alpha))
    return(mean(alphas > 0.5))
    
    
  },
  compute_paths = function(){
    NULL
  },
  density_active_path = function(){
    NA
  }
)

model_conv <- LBBNN_ConvNet(conv_layer_1, conv_layer_2, linear_layer_1,
                       linear_layer_2, device)
model_conv$to(device = device)
```

## Train and validate the model

These functions work the same as with feed-forward architectures. Training can be accelerated on GPU where available. 
```{r, eval = has_torch}
train_lbbnn(epochs = 2, LBBNN = model_conv, lr = 0.01, train_dl = train_loader,
            device = device)
validate_lbbnn(model_conv, num_samples = 2, test_dl = train_loader, 
               device = device)
```

---
title: "classification: gallstone dataset"
output: rmarkdown::html_vignette
params:
  eval: true
vignette: >
  %\VignetteIndexEntry{classification: gallstone dataset}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(LBBNN)
has_torch <- requireNamespace("torch", quietly = TRUE) &&
            torch::torch_is_installed()
```

## Prepare data

The gallstone dataset consists of a mix of demographic, bioimpedance and laboratory features. It is a mix of binary and continuous features, with 40 in total after pre-processing the dataset. It is relatively small, with only 319 total samples, where we use 223 of those for training and the remaining 96 for validation. This makes it potentially sensitive to initializations, or other hyperparameters such as the number of epochs to train.
```{r, eval = has_torch}
torch::torch_manual_seed(42)
loaders_gs <- get_dataloaders(gallstone_dataset, train_proportion = 0.70,
                              train_batch_size = 223, test_batch_size = 96,
                              standardize = TRUE, seed = 42)
train_loader_gs <- loaders_gs$train_loader
test_loader_gs <- loaders_gs$test_loader
```

## Define model

This 4 hidden layer model with normalizing flows has over 30000 parameters, which
can be seen when using the print function.
```{r, eval = has_torch}
problem <- "binary classification"
sizes <- c(40, 16, 16, 16, 16, 1)
inclusion_priors <- c(0.5, 0.5, 0.5, 0.5, 0.5) 
stds <- c(1, 1, 1, 1, 1) 
inclusion_inits <- 'polarized_dense'
device <- "cpu"
model_gs <- lbbnn_net(problem_type = problem, sizes = sizes,
                     prior = inclusion_priors, flow = TRUE,
                     dims = c(10, 10, 10, 10), 
                     inclusion_inits = inclusion_inits,
                     input_skip = TRUE, std = stds, device = device)
#print(model_gs)
```

## Train and validate

```{r, eval = FALSE}
train_lbbnn(epochs = 1000, LBBNN = model_gs,
            lr = 0.005, train_dl = train_loader_gs, device = device,
            verbose = FALSE)

validate_lbbnn(LBBNN = model_gs, num_samples = 10, test_dl = test_loader_gs,
              device = device)

```

The above results in a very sparse model, where accuracy around 73%. Instead, training can be 
terminated at a specified minimum density using the argument "min_density" as follows:

```{r,eval = FALSE}
torch::torch_manual_seed(42)
model_2 <- lbbnn_net(problem_type = problem, sizes = sizes,
                     prior = inclusion_priors, flow = TRUE,
                     dims = c(10, 10, 10, 10), 
                     inclusion_inits = inclusion_inits,
                     input_skip = TRUE, std = stds, device = device)

train_lbbnn(epochs = 1000, LBBNN = model_2,
            lr = 0.005, train_dl = train_loader_gs, device = device,
            verbose = FALSE, min_density = 0.1)

validate_lbbnn(LBBNN = model_2, num_samples = 10, test_dl = test_loader_gs,
              device = device)
```
Accuracy is now improved to around 79%.

---
title: "variable selection: linear-data"
output: rmarkdown::html_vignette
params:
  eval: true
vignette: >
  %\VignetteIndexEntry{variable selection: linear-data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(LBBNN)
has_torch <- requireNamespace("torch", quietly = TRUE) &&
            torch::torch_is_installed()
```

## Generate data

We generate 1000 samples with 15 features, and make 3 of them relevant for the outcome. The goal is to see if we recover the 
correct variables, and their coefficients. The data preparation process follows the same steps as in the previous example.
Note that the dataset must be a data.frame object before applying the get_dataloaders function.
```{r, eval = has_torch}
i <- 1000
j <- 15
set.seed(42)
torch::torch_manual_seed(42)
X <- matrix(rnorm(i * j, mean = 0, sd = 1), ncol = j)
y_base <- c()
y_base <-  0.6 * X[, 1] - 0.4 * X[, 2] + 0.5 * X[, 3] + rnorm(n = i, sd = 0.1)
sim_data <- as.data.frame(X)
sim_data <- cbind(sim_data, y_base)
loaders <- get_dataloaders(sim_data, train_proportion = 0.9,
                           train_batch_size = 450, test_batch_size = 100,
                           standardize = FALSE)
train_loader <- loaders$train_loader
test_loader  <- loaders$test_loader
```

## Define hyperparameters and the model object

Here we must use the "regression" keyword as we have a continuous outcome. We use 
a relatively small network of 2 hidden layers with 5 neurons in each. 
```{r, eval = has_torch}
problem <- "regression"
sizes <- c(j, 5, 5, 1) 
incl_priors <- c(0.5, 0.5, 0.5) 
stds <- c(1, 1, 1)
incl_inits <- 'polarized'
device <- "cpu" 
model_linear <- lbbnn_net(problem_type = problem, sizes = sizes,
                              prior = incl_priors, inclusion_inits = incl_inits,
                              std = stds, input_skip = TRUE, flow = FALSE,
                              num_transforms = 2, dims = c(10, 10, 10),
                              raw_output = FALSE, custom_act = NULL,
                              link = NULL, nll = NULL,
                              bias_inclusion_prob = FALSE, device = device)
```

## Train and validate the model

```{r, eval = has_torch}
train_lbbnn(epochs = 50, LBBNN = model_linear,
            lr = 0.1, train_dl = train_loader, device = device, verbose = FALSE)
validate_lbbnn(LBBNN = model_linear, num_samples = 2, test_dl = test_loader,
              device = device)
```

## Inspect the results using coef

Here we have given the dataset as an argument instead of picking a specific sample,
where the argument num_data controls how many samples from the dataset we use. Using inds = NULL,
we select random samples from the dataset. Explanations are then averaged across these samples.
```{r, eval = has_torch}
coef(model_linear, dataset = train_loader, inds = NULL,
     output_neuron = 1, num_data = 5, num_samples = 10)
```
The coefficients match the data generating process.

## Global structure

Using plot with type = 'global' gives us the global explanations, i.e. the features that affect predictions 
in general.
```{r,,fig.width=6, fig.height=6, eval = has_torch}
plot(model_linear, type = "global", vertex_size = 7,
     edge_width = 0.4, label_size = 0.4)
```

Everything has been pruned away aside from the linear connections from the relevant features.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/overwrite_functions.R
\name{predict.lbbnn_net}
\alias{predict.lbbnn_net}
\title{Obtain predictions from the posterior of an \code{LBBNN model}}
\usage{
\method{predict}{lbbnn_net}(
  object,
  newdata,
  mpm = FALSE,
  draws = 10,
  device = "cpu",
  link = NULL,
  ...
)
}
\arguments{
\item{object}{A trained \code{lbbnn_net} object}

\item{newdata}{A \code{torch::dataloader}
object containing the data with which to predict.}

\item{mpm}{logical, whether to use the median probability model.}

\item{draws}{integer, the number of samples to draw from the posterior.}

\item{device}{character, device for computation (default = \code{"cpu"}).}

\item{link}{Optional link function to apply to the network output.}

\item{...}{further arguments passed to or from other methods.}
}
\value{
A \code{torch::torch_tensor}  of shape \code{(draws,N,C)}
where \code{N} is the number of samples in \code{newdata},
and \code{C} the number of outputs.
}
\description{
Draw from the posterior of a trained \code{lbbnn_net} object.
}
\examples{
\donttest{
if (torch_available()) {
x<-torch::torch_randn(3,2)
b <- torch::torch_rand(2)
y <- torch::torch_matmul(x,b)
train_data <- torch::tensor_dataset(x,y)
train_loader <- torch::dataloader(train_data,batch_size = 3,shuffle=FALSE)
problem<-'regression'
sizes <- c(2,1,1)
inclusion_priors <-c(0.9,0.2)
inclusion_inits <- matrix(rep(c(-10,10),2),nrow = 2,ncol = 2)
stds <- c(1.0,1.0)
model <- lbbnn_net(problem,sizes,inclusion_priors,stds,inclusion_inits,
flow = FALSE,input_skip = TRUE)
train_lbbnn(epochs = 1,LBBNN = model, lr = 0.01,train_dl = train_loader)
predict(model,mpm = FALSE,newdata = train_loader,draws = 1)}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/local_explanations.R
\name{get_local_explanations_gradient}
\alias{get_local_explanations_gradient}
\title{Get gradient based local explanations for input-skip LBBNNs.}
\usage{
get_local_explanations_gradient(
  model,
  input_data,
  num_samples = 1,
  magnitude = TRUE,
  include_potential_contribution = FALSE,
  device = "cpu"
)
}
\arguments{
\item{model}{A \code{LBBNN_Net} with input-skip}

\item{input_data}{The data to be explained (one sample).}

\item{num_samples}{integer, samples to use to produce credible intervals.}

\item{magnitude}{If TRUE, only return explanations.
If FALSE, multiply by input values.}

\item{include_potential_contribution}{IF TRUE, If covariate=0,
we assume that the contribution is negative
(good/bad that it is not included) if FALSE, just removes zero covariates.}

\item{device}{character, the device to be trained on.
Default is 'cpu', can be 'mps' or 'gpu'.}
}
\value{
A list with the following elements:
\describe{
\item{explanations}{A \code{torch::tensor}
of shape (num_samples, p, num_classes).}
\item{p}{integer, the number of input features.}
\item{predictions}{A \code{torch::tensor}
of shape (num_samples, num_classes).}
}
}
\description{
Works by computing the gradient wrt to input, given we have
relu activation functions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/local_explanations.R
\name{quants}
\alias{quants}
\title{Function to obtain empirical 95\% confidence interval, including the mean}
\usage{
quants(x)
}
\arguments{
\item{x}{numeric vector whose sample quantiles is desired.}
}
\value{
The quantiles in addition to the mean.
}
\description{
Using the built in quantile function to return 95\% confidence interval
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/overwrite_functions.R
\name{coef.lbbnn_net}
\alias{coef.lbbnn_net}
\title{Get model coefficients (local explanations) of \code{lbbnn_net} object}
\usage{
\method{coef}{lbbnn_net}(
  object,
  dataset,
  inds = NULL,
  output_neuron = 1,
  num_data = 1,
  num_samples = 10,
  ...
)
}
\arguments{
\item{object}{an object of class \code{lbbnn_net}.}

\item{dataset}{Either a \code{torch::dataloader} object,
or a \code{torch::torch_tensor} object.
The former is assumed to be the same \code{torch::dataloader}
used for training or testing.
The latter can be any user-defined data.}

\item{inds}{Optional integer vector of row indices in the dataset to compute explanations for.}

\item{output_neuron}{integer, which output neuron to explain (default = 1).}

\item{num_data}{integer, if no indices are chosen,
the first \code{num_data} of \code{dataset} are automatically used for explanations.}

\item{num_samples}{integer, how many samples to use for model averaging
when sampling the weights in the active paths.}

\item{...}{further arguments passed to or from other methods.}
}
\value{
A data frame with rows corresponding to input variables
and the following columns:
\itemize{
\item \code{lower}: lower bound of the 95\% confidence interval
\item \code{mean}: mean contribution of the variable
\item \code{upper}: upper bound of the 95\% confidence interval
}
}
\description{
Given an input sample x_1,... x_j (with j the number of inputs,
the local explanation is found by considering active paths.
If relu activation functions are assumed, each path is a piecewise
linear function, so the contribution for x_j is just the sum of the weights
associated with the paths connecting x_j to the output.
The contributions are found by taking the gradient wrt x.
}
\details{
\itemize{
\item If \code{num_data = 1}, confidence intervals are computed using
model averaging over \code{num_samples} weight samples.
\item If \code{num_data > 1}, confidence intervals are computed across
.         the mean explanations for each sample.
\item The output is a data frame with row names as input variables
(\code{x0}, \code{x1}, \code{x2}, ...) and columns giving mean
and 95\% confidence intervals for each variable.
}
}
\examples{
\donttest{
if (torch_available()) {
x<-torch::torch_randn(3,2)
b <- torch::torch_rand(2)
y <- torch::torch_matmul(x,b)
train_data <- torch::tensor_dataset(x,y)
train_loader <- torch::dataloader(train_data,batch_size = 3,shuffle=FALSE)
problem<-'regression'
sizes <- c(2,1,1)
inclusion_priors <-c(0.9,0.2)
inclusion_inits <- matrix(rep(c(-10,10),2),nrow = 2,ncol = 2)
stds <- c(1.0,1.0)
model <- lbbnn_net(problem,sizes,inclusion_priors,stds,inclusion_inits,flow = FALSE,
input_skip = TRUE)
train_lbbnn(epochs = 1,LBBNN = model, lr = 0.01,train_dl = train_loader)
coef(model,dataset = x, num_data = 1)}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Layers.R
\name{lbbnn_linear}
\alias{lbbnn_linear}
\title{Class to generate an LBBNN feed forward layer}
\usage{
lbbnn_linear(
  in_features,
  out_features,
  prior_inclusion,
  standard_prior,
  density_init,
  flow = FALSE,
  num_transforms = 2,
  hidden_dims = c(200, 200),
  device = "cpu",
  bias_inclusion_prob = FALSE,
  conv_net = FALSE
)
}
\arguments{
\item{in_features}{integer, number of input neurons.}

\item{out_features}{integer, number of output neurons.}

\item{prior_inclusion}{numeric scalar, prior inclusion probability
for each weight and bias in the layer.}

\item{standard_prior}{numeric scalar, prior standard deviation
for weights and biases in each layer.}

\item{density_init}{A numeric of size 2,
used to initialize the inclusion parameters, one for each layer.}

\item{flow}{logical, whether to use normalizing flows}

\item{num_transforms}{integer, number of transformations for \code{flow}.
Default is 2.}

\item{hidden_dims}{numeric vector, dimension of the hidden layer(s)
in the neural networks of the RNVP transform.}

\item{device}{The device to be used. Default is CPU.}

\item{bias_inclusion_prob}{logical, determines whether the bias
should be as associated with inclusion probabilities.}

\item{conv_net}{logical, whether the layer is used in a convolutional net.}
}
\value{
A \code{torch::nn_module} object,
representing a fully connected LBBNN layer.
The module has the following methods:
\itemize{
\item \code{forward(input, MPM = FALSE)}: Computes activation
(using the LRT at training time) of a batch of inputs.
\item \code{kl_div()}: Computes the KL-divergence.
\item \code{sample_z()}: Samples from the flow if \code{flow = TRUE},
in addition returns the log-determinant of the Jacobian
of the transformation.
}
}
\description{
This module implements a fully connected LBBNN layer.
It supports:
\itemize{
\item Prior inclusion probabilities for weights and biases in each layer.
\item Standard deviation priors for weights and biases in each layer.
\item Optional normalizing flows (RNVP) for a more flexible posterior.
\item Forward pass using either the full model,
or the Median Probability Model (MPM).
\item Computation of the KL-divergence.
}
}
\examples{
\donttest{
if (torch_available()) {
l1 <- lbbnn_linear(in_features = 10,out_features = 5,prior_inclusion = 0.25,
standard_prior = 1,density_init = c(0,1),flow = FALSE)
x <- torch::torch_rand(20,10,requires_grad = FALSE)
output <- l1(x,MPM = FALSE) #the forward pass, output has shape (20,5)
print(l1$kl_div()$item())} #compute KL-divergence after the forward pass
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/overwrite_functions.R
\name{get_input_inclusions}
\alias{get_input_inclusions}
\title{Function that checks how many times inputs are included,
and from which layer. Used in summary function.}
\usage{
get_input_inclusions(model)
}
\arguments{
\item{model}{An instance of \code{lbbnn_net} where \code{input_skip = TRUE}.}
}
\value{
A matrix of shape (p, L-1) where p is the number of input variables
and L the total number of layers (including input and output),
with each element being 1 if the variable is included or 0 if not included.
}
\description{
Useful when the number of inputs and/or hidden neurons are very
large, and direct visualization of the network is difficult.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{gallstone_dataset}
\alias{gallstone_dataset}
\title{Gallstone Dataset}
\format{
This dataset has 319 rows and  38 columns.
}
\source{
\url{https://pmc.ncbi.nlm.nih.gov/articles/PMC11309733/#T2}
}
\usage{
gallstone_dataset
}
\description{
Taken from the UCI machine learning repository. The task is
to classify whether the patient had gallstones or not. It contains a mix
of demographic data and bioimpedance data.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/overwrite_functions.R
\name{residuals.lbbnn_net}
\alias{residuals.lbbnn_net}
\title{Residuals from LBBNN fit}
\usage{
\method{residuals}{lbbnn_net}(object, type = c("response"), ...)
}
\arguments{
\item{object}{An object of class \code{lbbnn_net}.}

\item{type}{Only 'response' is implemented i.e. y_true - y_predicted.}

\item{...}{further arguments passed to or from other methods.}
}
\value{
A numeric vector of residuals (\code{y_true - y_predicted})
}
\description{
Residuals from an object of the \code{lbbnn_net} class.
}
\examples{
\donttest{
if (torch_available()) {
x<-torch::torch_randn(3,2)
b <- torch::torch_rand(2)
y <- torch::torch_matmul(x,b)
train_data <- torch::tensor_dataset(x,y)
train_loader <- torch::dataloader(train_data,batch_size = 3,shuffle=FALSE)
problem<-'regression'
sizes <- c(2,1,1)
inclusion_priors <-c(0.9,0.2)
inclusion_inits <- matrix(rep(c(-10,10),2),nrow = 2,ncol = 2)
stds <- c(1.0,1.0)
model <- lbbnn_net(problem,sizes, inclusion_priors, stds ,inclusion_inits,
flow = FALSE, input_skip = TRUE)
train_lbbnn(epochs = 1,LBBNN = model, lr = 0.01,train_dl = train_loader)
residuals(model)}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FLOW.R
\name{normalizing_flow}
\alias{normalizing_flow}
\title{Class to generate a normalizing flow}
\usage{
normalizing_flow(input_dim, transform_type, num_transforms)
}
\arguments{
\item{input_dim}{numeric vector, the dimensionality of each layer.
The first item is the input vector size.}

\item{transform_type}{Transformation type. Currently RNVP is implemented.}

\item{num_transforms}{integer, how many layers of transformations to include.}
}
\value{
A \code{torch::nn_module} object representing the normalizing flow.
The module provides:
\describe{
\item{\code{forward(z)}}{
Applies all flow transformation layers to the input tensor \code{z}.
Returns a named list containing:
\describe{
\item{\code{z}}{
A \code{torch_tensor} containing the transformed version of the input,
with the same shape as \code{z}.
}
\item{\code{logdet}}{
A scalar \code{torch_tensor} equal to the sum of the log-determinants of
all transformation layers.
}
}
}
}
}
\description{
Used in\code{LBBNN_Net} when the argument \code{flow = TRUE}.
Contains a \code{torch::nn_module} where the initial vector gets transformed
through all the layers in the module.
Also computes the log-determinant of the Jacobian for the entire
transformation, the sum of the log-determinants of the independent layers.
}
\examples{
\donttest{
if (torch_available()) {
flow <- normalizing_flow(c(2,5,5), transform_type='RNVP', num_transforms = 3)
flow$to(device = 'cpu')
x <- torch::torch_rand(2, device = 'cpu')
output <- flow(x)
z_out <- output$z
print(dim(z_out))
log_det <- output$logdet
print(log_det)}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LBBNN_Model.R
\name{lbbnn_net}
\alias{lbbnn_net}
\title{Feed-forward Latent Binary Bayesian Neural Network (LBBNN)}
\usage{
lbbnn_net(
  problem_type,
  sizes,
  prior,
  std,
  inclusion_inits,
  input_skip = FALSE,
  flow = FALSE,
  num_transforms = 2,
  dims = c(200, 200),
  device = "cpu",
  raw_output = FALSE,
  custom_act = NULL,
  link = NULL,
  nll = NULL,
  bias_inclusion_prob = FALSE
)
}
\arguments{
\item{problem_type}{character, one of:
\code{'binary classification'}, \code{'multiclass classification'},
\code{'regression'}, or \code{'custom'}.}

\item{sizes}{Integer vector specifying the layer sizes of the network.
The first element is the input size,
the last is the output size,
and the intermediate integers represent hidden layers.}

\item{prior}{numeric vector of prior inclusion probabilities for
each weight matrix. length must be \code{length(sizes) - 1}.}

\item{std}{numeric vector of prior standard deviation for each weight matrix.
length must be \code{length(sizes) - 1}.}

\item{inclusion_inits}{numeric matrix of shape (2, number of weight matrices)
specifying the lower and upper bounds for initializations
of the inclusion parameters.}

\item{input_skip}{logical, whether to include input_skip.}

\item{flow}{logical, whether to use normalizing flows.}

\item{num_transforms}{integer, how many transformations to use in the flow.}

\item{dims}{numeric vector, hidden dimension for the neural network
in the RNVP transform.}

\item{device}{the device to be trained on. Can be 'cpu', 'gpu' or 'mps'.
Default is cpu.}

\item{raw_output}{logical, whether the network skips the last sigmoid/softmax
layer to compute local explanations.}

\item{custom_act}{Allows the user to submit their own customized
activation function.}

\item{link}{User can define their own link function (not implemented).}

\item{nll}{User can define their own likelihood function (not implemented).}

\item{bias_inclusion_prob}{logical, determines whether the bias should
be as associated with inclusion probabilities.}
}
\value{
A \code{torch::nn_module} object representing the LBBNN.
It includes the following methods:
\itemize{
\item \code{forward(x, MPM = FALSE)}: Performs a forward pass
through the whole network.
\item \code{kl_div()}: Returns the KL divergence of the network.
\item \code{density()}: Returns the density of the whole network,
i.e. the proportion of weights
with inclusion probabilities greater than 0.5.
\item \code{compute_paths()}: Computes active paths through the network
without input-skip.
\item \code{compute_paths_input_skip()}: Computes active paths with
input-skip enabled.
\item \code{density_active_path()}: Returns network density after
removing inactive paths.
}
}
\description{
Each layer is defined by \code{lbbnn_linear}.
For example, \code{sizes = c(20, 200, 200, 5)} generates a network with:
\itemize{
\item 20 input features,
\item two hidden layers of 200 neurons each,
\item an output layer with 5 neurons.
}
}
\examples{
\donttest{
if (torch_available()) {
layers <- c(10,2,5)
alpha <- c(0.3,0.9)
stds <- c(1.0,1.0)
inclusion_inits <- matrix(rep(c(-10,10),2),nrow = 2,ncol = 2)
prob <- 'multiclass classification'
net <- lbbnn_net(problem_type = prob, sizes = layers, prior = alpha,
std = stds, inclusion_inits = inclusion_inits,input_skip = FALSE,
flow = FALSE,device = 'cpu')
x <- torch::torch_rand(20,10,requires_grad = FALSE)
output <- net(x)
net$kl_div()$item()
net$density()}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RNVP.R
\name{mlp}
\alias{mlp}
\title{Multi layer-perceptron}
\usage{
mlp(hidden_sizes, device = "cpu")
}
\value{
Returns a \code{torch::nn_module} representing the mlp.
The module has the following method:
\describe{
\item{\code{forward(x)}}{
Applies each linear layer in \code{hidden_sizes} followed by a LeakyReLU
activation (except after
the final layer). Returns a \code{torch::torch_tensor}
whose last dimension equals the
last element of \code{hidden_sizes}.
}
}
}
\description{
Generate a multi-layer perceptron, used in RNVP transforms.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Train_validate.R
\name{validate_lbbnn}
\alias{validate_lbbnn}
\title{Validate a trained LBBNN model.}
\usage{
validate_lbbnn(LBBNN, num_samples, test_dl, device = "cpu")
}
\arguments{
\item{LBBNN}{An instance of a trained \code{lbbnn_net} to be validated.}

\item{num_samples}{integer, the number of samples from the variational
posterior to be used for model averaging.}

\item{test_dl}{An instance of \code{torch::dataloader},
containing the validation data.}

\item{device}{The device to perform validation on.
Default is 'cpu'; other options include 'gpu' and 'mps'.}
}
\value{
A list containing the following elements:
\describe{
\item{accuracy_full_model}{Classification accuracy of the
full (dense) model (if classification).}
\item{accuracy_sparse}{Classification accuracy using only
weights in active paths (if classification).}
\item{validation_error}{Root mean squared error
for the full model (if regression).}
\item{validation_error_sparse}{Root mean squared error
using only weights in active paths (if regression).}
\item{density}{Proportion of weights with posterior
inclusion probability > 0.5 in the whole network.}
\item{density_active_path}{Proportion of weights
.    with inclusion probability > 0.5 after removing weights not in
.    active paths.}
}
}
\description{
Computes metrics on a validation dataset without computing gradients.
Supports model averaging (recommended) by
sampling from the variational posterior (\code{num_samples} > 1)
to improve predictions. Returns metrics for both the full model
and the sparse model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Layers.R
\name{std_prior}
\alias{std_prior}
\title{Generate prior standard deviation for weights and biases of either
linear or convolutional layers.}
\usage{
std_prior(x)
}
\arguments{
\item{x}{A number greater than 0.}
}
\value{
a numeric to be added to either (out_shape,in_shape)
in case of linear layers or (out_channels,in_channels,kernel0,kernel1)
in case of convolutional layers.
}
\description{
A function to generate prior standard deviations for
each weight and bias in an LBBNN layer.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Layers.R
\name{lbbnn_conv2d}
\alias{lbbnn_conv2d}
\title{Class to generate an LBBNN convolutional layer.}
\usage{
lbbnn_conv2d(
  in_channels,
  out_channels,
  kernel_size,
  prior_inclusion,
  standard_prior,
  density_init,
  flow = FALSE,
  num_transforms = 2,
  hidden_dims = c(200, 200),
  device = "cpu"
)
}
\arguments{
\item{in_channels}{integer, number of input channels.}

\item{out_channels}{integer, number of output channels.}

\item{kernel_size}{size of the convolving kernel.}

\item{prior_inclusion}{numeric scalar, prior inclusion probability for
each weight and bias in the layer.}

\item{standard_prior}{numeric scalar, prior standard deviation for
weights and biases in each layer.}

\item{density_init}{A numeric of size 2, used to initialize the
inclusion parameters, one for each layer.}

\item{flow}{logical, whether to use normalizing flows}

\item{num_transforms}{integer, number of transformations for \code{flow}.
Default is 2.}

\item{hidden_dims}{numeric vector, dimension of the hidden layer(s)
in the neural networks of the RNVP transform.}

\item{device}{The device to be used. Default is CPU.}
}
\value{
A \code{torch::nn_module} object representing a convolutional
LBBNN layer.
The module has the following methods:
\itemize{
\item \code{forward(input, MPM = FALSE)}: Computes activation
(using the LRT at training time) of a batch of inputs.
\item \code{kl_div()}: Computes the KL-divergence.
\item \code{sample_z()}: Samples from the flow if \code{flow = TRUE},
in addition returns the log-determinant of the Jacobian
of the transformation.
}
}
\description{
It supports:
\itemize{
\item Prior inclusion probabilities for weights and biases in each layer.
\item Standard deviation priors for weights and biases in each layer.
\item Optional normalizing flows (RNVP) for a more flexible posterior.
\item Forward pass using either the full model or the
Median Probability Model (MPM).
\item Computation of the KL-divergence.
}
}
\examples{
\donttest{
if (torch_available()) {
layer <- lbbnn_conv2d(in_channels = 1,out_channels = 32,kernel_size = c(3,3),
prior_inclusion = 0.2,standard_prior = 1,density_init = c(0,1),device = 'cpu')
x <-torch::torch_randn(100,1,28,28)
out <-layer(x)
print(dim(out))}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_graphs.R
\name{get_adj_mats}
\alias{get_adj_mats}
\title{Obtain adjacency matrices for \code{igraph} plotting}
\usage{
get_adj_mats(model)
}
\arguments{
\item{model}{An instance of \code{lbbnn_net} with input_skip enabled.}
}
\value{
A list of adjacency matrices, one for each hidden layer and output.
}
\description{
Takes the alpha active path matrices for each layer of the
LBBNN and converts them to adjacency matrices so that they can be plotted
with igraph.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/custom_activation_functions.R
\name{custom_activation}
\alias{custom_activation}
\title{Generate a custom activation function.}
\usage{
custom_activation()
}
\value{
Returns a `\code{torch::nn_module} that can be used in an \code{LBBNN_Net}
}
\description{
The first 3 entries are customized
in order to see if we can learn that structure. The rest will be relu as usual
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RNVP.R
\name{rnvp_layer}
\alias{rnvp_layer}
\title{Single RNVP transform layer.}
\usage{
rnvp_layer(hidden_sizes, device = "cpu")
}
\arguments{
\item{hidden_sizes}{A vector of integers.
The first is the dimensionality of the vector,
to be transformed by RNVP. The subsequent are hidden dimensions in the mlp.}

\item{device}{The device to be used. Default is CPU.}
}
\value{
A \code{torch::nn_module} object representing a single RNVP layer.
The module has the following methods:
\describe{
\item{\code{forward(z)}}{Applies the RNVP transformation.
Returns a \code{torch::torch_tensor} with the
same shape as z.}
\item{\code{log_det()}}{A scalar \code{torch::torch_tensor}
giving the log-determinant of the Jacobian of the transformation.}
}
}
\description{
Affine half flow aka Real Non-Volume Preserving
(x = z * exp(s) + t), where a randomly selected half z1 of the dimensions
in z are transformed as an Affine function of the other half z2,
i.e. scaled by s(z2) and shifted by t(z2).
From "Density estimation using Real NVP", Dinh et al. (May 2016)
https://arxiv.org/abs/1605.08803
This implementation uses the numerically stable updates introduced by IAF:
https://arxiv.org/abs/1606.04934
}
\examples{
\donttest{
if (torch_available()) {
z <- torch::torch_rand(200)
layer <- rnvp_layer(c(200,50,100))
out <- layer(z)
print(dim(out))
print(layer$log_det())}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LBBNN-package.R
\docType{package}
\name{LBBNN-package}
\alias{LBBNN}
\alias{LBBNN-package}
\title{LBBNN: Latent Binary Bayesian Neural Networks Using 'torch'}
\description{
Latent binary Bayesian neural networks (LBBNNs) are implemented using 'torch', an R interface to the LibTorch backend. Supports mean-field variational inference as well as flexible variational posteriors using normalizing flows. The standard LBBNN implementation follows Hubin and Storvik (2024) \doi{10.3390/math12060788}, using the local reparametrization trick as in Skaaret-Lund et al. (2024) \url{https://openreview.net/pdf?id=d6kqUKzG3V}. Input-skip connections are also supported, as described in Høyheim et al. (2025) \doi{10.48550/arXiv.2503.10496}.
}
\author{
\strong{Maintainer}: Lars Skaaret-Lund \email{lars.skaaret-lund@nmbu.no}

Authors:
\itemize{
  \item Aliaksandr Hubin \email{aliaksandr.hubin@nmbu.no}
  \item Eirik Høyheim \email{eirik.hoyheim@ffi.no}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/overwrite_functions.R
\name{plot.lbbnn_net}
\alias{plot.lbbnn_net}
\title{Plot \code{lbbnn_net} objects}
\usage{
\method{plot}{lbbnn_net}(x, type = c("global", "local"), data = NULL, num_samples = 100, ...)
}
\arguments{
\item{x}{An instance of \code{lbbnn_net}.}

\item{type}{Either \code{"global"} or \code{"local"}.}

\item{data}{If local is chosen, one sample must be provided to obtain
the explanation. Must be a \code{torch::torch_tensor} of shape \code{(1,p)}.}

\item{num_samples}{integer, how many samples to use for model averaging
over the weights in case of local explanations.}

\item{...}{further arguments passed to or from other methods.}
}
\value{
No return value. Called for its side effects of producing a plot.
}
\description{
Given a trained \code{lbbnn_net} model, this function produces either:
\itemize{
\item \strong{Global plot}: a visualization of the network structure,
showing only the active paths.
\item \strong{Local explanation}: a plot of the local
explanation for a single input sample, including error bars obtained
from Monte Carlo sampling of the network weights.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/local_explanations.R
\name{plot_local_explanations_gradient}
\alias{plot_local_explanations_gradient}
\title{Plot the gradient based local explanations for one sample.}
\usage{
plot_local_explanations_gradient(
  model,
  input_data,
  num_samples,
  device = "cpu",
  save_svg = NULL
)
}
\arguments{
\item{model}{An instance of \code{LBBNN_Net} with input-skip enabled.}

\item{input_data}{The data to be explained (one sample).}

\item{num_samples}{integer, samples to use to produce credible intervals.}

\item{device}{character, the device to be trained on. Default is cpu.
Can be 'mps' or 'gpu'.}

\item{save_svg}{the path where the plot will be saved as svg,
if save_svg is not NULL.}
}
\value{
This function produces plots as a side
effect and does not return a value.
}
\description{
Plots the contribution of each covariate,
and the prediction, with error bars.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Layers.R
\name{alpha_prior}
\alias{alpha_prior}
\title{Generate prior inclusion probabilities for the weights of a
LBBNN layer (linear or convolutional).}
\usage{
alpha_prior(x)
}
\arguments{
\item{x}{A number between 0 and 1.}
}
\value{
a numeric to be added to either (out_shape,in_shape)
in case of linear layers, or (out_channels,in_channels,kernel0,kernel1)
in case of convolutional layers.
}
\description{
A function to generate prior probability values for
each weight in an LBBNN layer.
The same probability is applied to all weights within a layer.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_graphs.R
\name{plot_active_paths}
\alias{plot_active_paths}
\title{Function to plot an input skip structure after removing weights
in non-active paths.}
\usage{
plot_active_paths(
  model,
  layer_spacing = 1,
  neuron_spacing = 1,
  vertex_size = 10,
  label_size = 0.5,
  edge_width = 0.5,
  save_svg = NULL
)
}
\arguments{
\item{model}{A trained \code{lbbnn_net} model with input_skip enabled.}

\item{layer_spacing}{numeric, spacing in between layers.}

\item{neuron_spacing}{numeric, spacing between neurons within a layer.}

\item{vertex_size}{numeric, size of the neurons.}

\item{label_size}{numeric, size of the text within neurons.}

\item{edge_width}{numeric, width of the edges connecting neurons.}

\item{save_svg}{the path where the plot will be saved.}
}
\value{
This function produces plots as a side effect and does not return a value.
}
\description{
Uses igraph to plot.
}
\examples{
\donttest{
if (torch_available()) {
sizes <- c(2,3,3,2)
problem <- 'multiclass classification'
inclusion_priors <- c(0.1,0.1,0.1)
std_priors <- c(1.0,1.0,1.0)
inclusion_inits <- matrix(rep(c(-10,10),3), nrow = 2, ncol = 3)
device <- 'cpu'
torch::torch_manual_seed(0)
model <- lbbnn_net(problem_type = problem, sizes = sizes,
                 prior = inclusion_priors, inclusion_inits = inclusion_inits,
                  input_skip = TRUE, std = std_priors, flow = FALSE,
                  num_transforms = 2, dims = c(200,200), device = device)
model$compute_paths_input_skip()
LBBNN:::plot_active_paths(model, 1, 1, 14, 1)}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_torch.R
\name{torch_available}
\alias{torch_available}
\title{Check if torch/libtorch is available for examples}
\usage{
torch_available()
}
\description{
Check if torch/libtorch is available for examples
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataloader.R
\name{get_dataloaders}
\alias{get_dataloaders}
\title{Wrapper around \code{torch::dataloader}}
\usage{
get_dataloaders(
  dataset,
  train_proportion,
  train_batch_size,
  test_batch_size,
  standardize = TRUE,
  shuffle_train = TRUE,
  shuffle_test = FALSE,
  seed = 1
)
}
\arguments{
\item{dataset}{A \code{data.frame}.
The last column is assumed to be the dependent variable.}

\item{train_proportion}{numeric, between 0 and 1.
Proportion of data to be used for training.}

\item{train_batch_size}{integer, samples per batch in the train dataloader.}

\item{test_batch_size}{integer, samples per batch in the test dataloader.}

\item{standardize}{logical, standardize input-features, default is TRUE.}

\item{shuffle_train}{logical, shuffle training data each epoch. default TRUE}

\item{shuffle_test}{logical, shuffle test data, default is FALSE.}

\item{seed}{integer. Used for reproducibility in the train/test split.}
}
\value{
A list containing:
\describe{
\item{train_loader}{A \code{torch::dataloader} for the training data.}
\item{test_loader}{A \code{torch::dataloader} for the test data.}
}
}
\description{
Avoids users having to manually define their own dataloaders.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/overwrite_functions.R
\name{print.lbbnn_net}
\alias{print.lbbnn_net}
\title{Print summary of an \code{lbbnn_net} object}
\usage{
\method{print}{lbbnn_net}(x, ...)
}
\arguments{
\item{x}{An object of class \code{lbbnn_net}.}

\item{...}{Further arguments passed to or from other methods.}
}
\value{
Invisibly returns the input \code{x}.
}
\description{
Provides a summary of a trained \code{lbbnn_net} object.
Includes the model type (input-skip or not), whether normalizing flows
are used, module and sub-module structure, number of trainable parameters,
and prior variance and inclusion probabilities for the weights.
}
\examples{
\donttest{
if (torch_available()) {
x<-torch::torch_randn(3,2)
b <- torch::torch_rand(2)
y <- torch::torch_matmul(x,b)
train_data <- torch::tensor_dataset(x,y)
train_loader <- torch::dataloader(train_data,batch_size = 3,shuffle=FALSE)
problem<-'regression'
sizes <- c(2,1,1)
inclusion_priors <-c(0.9,0.2)
inclusion_inits <- matrix(rep(c(-10,10),2),nrow = 2,ncol = 2)
stds <- c(1.0,1.0)
model <- lbbnn_net(problem,sizes,inclusion_priors,stds,inclusion_inits,
flow = FALSE, input_skip = TRUE)
print(model)}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Train_validate.R
\name{train_lbbnn}
\alias{train_lbbnn}
\title{Train an instance of \code{lbbnn_net}.}
\usage{
train_lbbnn(
  epochs,
  LBBNN,
  lr,
  train_dl,
  device = "cpu",
  scheduler = NULL,
  sch_step_size = NULL
)
}
\arguments{
\item{epochs}{integer, total number of epochs to train for, where one epoch
is a pass through the entire training dataset (all mini batches).}

\item{LBBNN}{An instance of  \code{lbbnn_net}, to be trained.}

\item{lr}{numeric, the learning rate to be used in the Adam optimizer.}

\item{train_dl}{An instance of \code{torch::dataloader}
consisting of a tensor dataset with features and targets.}

\item{device}{the device to be trained on. Default is 'cpu',
also accepts 'gpu' or 'mps'.}

\item{scheduler}{A torch learning rate scheduler object.
Can be used to decay learning rate for better convergence,
currently only supports 'step'.}

\item{sch_step_size}{Where to decay if using \code{torch::lr_step}.
E.g. 1000 means learning rate is decayed every 1000 epochs.}
}
\value{
a list containing the losses and accuracy (if classification)
and density for each epoch during training.
For comparisons sake we show the density with and without active paths.

A list with elements (returned invisibly):
\describe{
\item{accs}{Vector of accuracy per epoch (classification only).}
\item{loss}{Vector of average loss per epoch.}
\item{density}{Vector of network densities per epoch.}
}
}
\description{
Function that for each epoch iterates through each mini-batch,
computing the loss and using back-propagation to update network parameters.
}
\examples{
\donttest{
if (torch_available()) {
x<-torch::torch_randn(3,2)
b <- torch::torch_rand(2)
y <- torch::torch_matmul(x,b)
train_data <- torch::tensor_dataset(x,y)
train_loader <- torch::dataloader(train_data,batch_size = 3,shuffle=FALSE)
problem<-'regression'
sizes <- c(2,1,1)
inclusion_priors <-c(0.9,0.2)
inclusion_inits <- matrix(rep(c(-10,10),2),nrow = 2,ncol = 2)
stds <- c(1.0,1.0)
model <- lbbnn_net(problem,sizes,inclusion_priors,stds,inclusion_inits,
flow = FALSE)
output <- train_lbbnn(epochs = 1,LBBNN = model, lr = 0.01,
train_dl = train_loader)}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_graphs.R
\name{assign_names}
\alias{assign_names}
\title{Assign names to nodes.}
\usage{
assign_names(model)
}
\arguments{
\item{model}{A trained \code{lbbnn_net} model with input-skip.}
}
\value{
A list of adjacency matrices with the correct names.
}
\description{
Internal helper function to assign descriptive names to
nodes used for plotting.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/overwrite_functions.R
\name{summary.lbbnn_net}
\alias{summary.lbbnn_net}
\title{Summary of LBBNN fit}
\usage{
\method{summary}{lbbnn_net}(object, ...)
}
\arguments{
\item{object}{An object of class \code{lbbnn_net}.}

\item{...}{further arguments passed to or from other methods.}
}
\value{
A \code{data.frame} containing the above information.
The function prints a formatted summary to the console.
The returned \code{data.frame} is invisible.
}
\description{
Summary method for objects of the \code{lbbnn_net} class.
Only applies to objects trained with \code{input_skip = TRUE}.
}
\details{
The returned table combines two types of information:
\itemize{
\item Number of times each input variable is included
in the active paths from each layer
(obtained from \code{get_input_inclusions()}).
\item Average inclusion probabilities for each input variable from each layer,
.        including a final column showing the average across all layers.
}
}
\examples{
\donttest{
if (torch_available()) {
x<-torch::torch_randn(3,2)
b <- torch::torch_rand(2)
y <- torch::torch_matmul(x,b)
train_data <- torch::tensor_dataset(x,y)
train_loader <- torch::dataloader(train_data,batch_size = 3,shuffle=FALSE)
problem<-'regression'
sizes <- c(2,1,1)
inclusion_priors <-c(0.9,0.2)
inclusion_inits <- matrix(rep(c(-10,10),2),nrow = 2,ncol = 2)
stds <- c(1.0,1.0)
model <- lbbnn_net(problem, sizes, inclusion_priors, stds, inclusion_inits,
flow = FALSE, input_skip = TRUE)
train_lbbnn(epochs = 1,LBBNN = model, lr = 0.01,train_dl = train_loader)
summary(model)}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting_graphs.R
\name{assign_within_layer_pos}
\alias{assign_within_layer_pos}
\title{Function for plotting nodes in the network between two layers.}
\usage{
assign_within_layer_pos(N, N_u, input_positions, neuron_spacing)
}
\arguments{
\item{N}{integer, number of neurons in the first layer.}

\item{N_u}{integer, number of neurons in the second layer.}

\item{input_positions}{Positions of the neurons in the input layer.}

\item{neuron_spacing}{How much space between the neurons.}
}
\value{
Positions of the second layer.
}
\description{
Takes care of the three possible cases. Both layers have even
number of neurons, both layers have odd numbers, or one of each.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{raisin_dataset}
\alias{raisin_dataset}
\title{Raisins Dataset}
\format{
this data frame has 900 rows and the following 8 columns:
\describe{
\item{Area}{Number of pixels within the boundary}
\item{MajorAxisLength}{Length of the main axis}
\item{MinorAxisLength}{Length of the small axis}
\item{Eccentricity}{Measure of the eccentricity of the ellipse}
\item{ConvexArea}{The number of pixels of the
smallest convex shell of the region formed by the raisin grain}
\item{Extent}{Ratio of the region formed by the
raisin grain to the total pixels in the bounding box}
\item{Perimeter}{distance between the boundaries of the
raisin grain and the pixels around it}
\item{Class}{Kecimen or Besni raisin.}
}
}
\source{
\url{https://archive.ics.uci.edu/dataset/850/raisin}
}
\usage{
raisin_dataset
}
\description{
Ilkay Cinar, Murat Kokl and Sakir Tasdemi(2020)
provide a dataset consisting of 2 varieties of Turkish raisins, with 450
samples of each type. The dataset contains 7 morphological features,
extracted from images taken of the Raisins. The goal is to classify to one of
the two types of Raisins.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Layers.R
\name{density_initialization}
\alias{density_initialization}
\title{Generate initializations for the inclusion parameters.}
\usage{
density_initialization(lower, upper)
}
\arguments{
\item{lower}{numeric scalar.}

\item{upper}{numeric scalar, must be greater than}
}
\value{
A numeric vector of length 2: \code{c(lower,upper)}
}
\description{
Controls the initial density of the network, an important tuning
parameter. The initialization parameters are sampled from 1/(1+exp(-U)),
with U ~ Uniform(lower,upper).
}
\keyword{internal}
